"""Setup commands."""
import os
import shutil
import sys

from setuptools import Command

# --------------------------------------------------------------------------- #
# Extra Commands
# --------------------------------------------------------------------------- #


class BuildApiDocsCommand(Command):
    """Helper command to build the available api documents.

    This scans all the subdirectories under api and runs the
    build.py script underneath trying to build the api
    documentation for the given format.
    """

    description = "build all the project's api documents"
    user_options = []

    def initialize_options(self):
        """Initialize options setup."""
        if not os.path.exists('./build'):
            os.mkdir('./build')

    def finalize_options(self):
        """Finalize options teardown."""
        pass

    def run(self):
        """Run command."""
        old_cwd = os.getcwd()
        directories = (d for d in os.listdir('./doc/api') if not d.startswith('.'))
        for entry in directories:
            os.chdir('./doc/api/%s' % entry)
            os.system('python build.py')
            os.chdir(old_cwd)


class DeepCleanCommand(Command):
    """Helper command to return the directory to a completely clean state."""

    description = "clean everything that we don't want"
    user_options = []
    trash = ['build', 'dist', 'pymodbus.egg-info',
             os.path.join(os.path.join('doc', 'sphinx'), 'build'),
             ]

    def initialize_options(self):
        """Initialize options setup."""
        pass

    def finalize_options(self):
        pass

    def run(self):
        """Run command."""
        self._delete_pyc_files()
        self._delete_trash_dirs()

    def _delete_trash_dirs(self):
        """Remove all directories created in building."""
        self._delete_pyc_files()
        for directory in self.trash:
            if os.path.exists(directory):
                shutil.rmtree(directory)

    @staticmethod
    def _delete_pyc_files():
        """Remove all python cache files."""
        for root, dirs, files in os.walk('.'):
            for file in files:
                if file.endswith('.pyc'):
                    os.remove(os.path.join(root, file))


class LintCommand(Command):
    """Helper command to perform a lint scan of the sourcecode and return the results."""

    description = "perform a lint scan of the code"
    user_options = []

    def initialize_options(self):
        """Initialize options setup."""
        if not os.path.exists('./build'):
            os.mkdir('./build')

    def finalize_options(self):
        pass

    def run(self):
        """Run command."""
        scanners = [s for s in dir(self) if s.find('__try') >= 0]
        for scanner in scanners:
            if getattr(self, scanner)():
                break

    def _try_pyflakes(self):
        try:
            from pyflakes.scripts.pyflakes import main
            sys.argv = """pyflakes pymodbus""".split()
            main()
            return True
        except Exception:
            return False

    def _try_pychecker(self):
        try:
            import pychecker  # noqa: F401
            sys.argv = """pychecker pymodbus/*.py""".split()
            main()  # noqa: F821
            return True
        except Exception:
            return False

    def _try_pylint(self):
        try:
            import pylint  # noqa: F401
            sys.argv = """pylint pymodbus/*.py""".split()
            main()  # noqa: F821
            return True
        except Exception:
            return False


class Python3Command(Command):
    """Helper command to scan for potential python 3 errors.

    ./setup.py scan_2to3 > build/diffs_2to3 build/report_2to3
    """

    description = "perform 2to3 scan of the code"
    user_options = []
    directories = ['pymodbus', 'test', 'examples']

    def initialize_options(self):
        """Initialize options setup."""
        if not os.path.exists('./build'):
            os.mkdir('./build')

    def finalize_options(self):
        pass

    def run(self):
        """Run command"""
        self._run_python3()

    def _run_python3(self):
        try:
            from lib2to3.main import main
            sys.argv = ['2to3'] + self.directories
            main("lib2to3.fixes")
            return True
        except Exception:
            return False


class Pep8Command(Command):
    """Helper command to scan for potential pep8 violations."""

    description = "perform pep8 scan of the code"
    user_options = []
    directories = ['pymodbus']

    def initialize_options(self):
        """Initialize options setup"""
        if not os.path.exists('./build'):
            os.mkdir('./build')

    def finalize_options(self):
        pass

    def run(self):
        """Run command."""
        self._run_pep8()

    def _run_pep8(self):
        try:
            from pep8 import _main as main
            sys.argv = """pep8 --repeat --count --statistics
            """.split() + self.directories
            main()
            return True
        except Exception:
            return False

# --------------------------------------------------------------------------- #
# Command Configuration
# --------------------------------------------------------------------------- #


command_classes = {
    'deep_clean': DeepCleanCommand,
    'build_apidocs': BuildApiDocsCommand,
    'lint': LintCommand,
    'scan_2to3': Python3Command,
    'pep8': Pep8Command,
}

# --------------------------------------------------------------------------- #
# Export command list
# --------------------------------------------------------------------------- #
__all__ = ['command_classes']



"""Client.

import external classes, to make them easier to use:
"""
from pymodbus.client.serial import AsyncModbusSerialClient, ModbusSerialClient
from pymodbus.client.tcp import AsyncModbusTcpClient, ModbusTcpClient
from pymodbus.client.tls import AsyncModbusTlsClient, ModbusTlsClient
from pymodbus.client.udp import AsyncModbusUdpClient, ModbusUdpClient
from pymodbus.client.base import ModbusBaseClient


# ---------------------------------------------------------------------------#
#  Exported symbols
# ---------------------------------------------------------------------------#
__all__ = [
    "AsyncModbusSerialClient",
    "AsyncModbusTcpClient",
    "AsyncModbusTlsClient",
    "AsyncModbusUdpClient",
    "ModbusBaseClient",
    "ModbusSerialClient",
    "ModbusTcpClient",
    "ModbusTlsClient",
    "ModbusUdpClient",
]



#!/usr/bin/env python3
"""Pymodbus Asynchronous Client Examples.

The following is an example of how to use the asynchronous serial modbus
client implementation from pymodbus using tornado.
"""
import functools

# ---------------------------------------------------------------------------#
# import needed libraries
# ---------------------------------------------------------------------------#
import logging

from tornado.ioloop import IOLoop

from pymodbus.client.asynchronous import schedulers
from pymodbus.client.asynchronous.serial import AsyncModbusSerialClient


# ---------------------------------------------------------------------------#
# choose the requested modbus protocol
# ---------------------------------------------------------------------------#


# ---------------------------------------------------------------------------#
# configure the client logging
# ---------------------------------------------------------------------------#

FORMAT = (
    "%(asctime)-15s %(threadName)-15s"
    " %(levelname)-8s %(module)-15s:%(lineno)-8s %(message)s"
)
log = logging.getLogger()
log.setLevel(logging.DEBUG)

UNIT = 0x01

# ---------------------------------------------------------------------------#
# helper method to test deferred callbacks
# ---------------------------------------------------------------------------#


def dassert(future, callback):
    """Dassert."""

    def _assertor(value):
        # by pass assertion, an error here stops the write callbacks
        assert value  # nosec

    def on_done(f_trans):
        if exc := f_trans.exception():
            log.debug(exc)
            return _assertor(False)

        return _assertor(callback(f_trans.result()))

    future.add_done_callback(on_done)


def _print(value):
    """Print."""
    if hasattr(value, "bits"):
        result = value.bits
    elif hasattr(value, "registers"):
        result = value.registers
    else:
        log.error(value)
        return None
    txt = f"Printing : -- {result}"
    log.info(txt)
    return result


# ---------------------------------------------------------------------------#
# example requests
# ---------------------------------------------------------------------------#
# simply call the methods that you would like to use. An example session
# is displayed below along with some assert checks. Note that unlike the
# synchronous version of the client, the asynchronous version returns
# deferreds which can be thought of as a handle to the callback to send
# the result of the operation.  We are handling the result using the
# deferred assert helper(dassert).
# ---------------------------------------------------------------------------#


def begin_asynchronous_test(client, protocol):
    """Begin async test."""
    rq = client.write_coil(1, True, unit=UNIT)
    rr = client.read_coils(1, 1, unit=UNIT)
    dassert(rq, lambda r: r.function_code < 0x80)  # test for no error
    dassert(rr, _print)  # test the expected value

    rq = client.write_coils(1, [False] * 8, unit=UNIT)
    rr = client.read_coils(1, 8, unit=UNIT)
    dassert(rq, lambda r: r.function_code < 0x80)  # test for no error
    dassert(rr, _print)  # test the expected value

    rq = client.write_coils(1, [False] * 8, unit=UNIT)
    rr = client.read_discrete_inputs(1, 8, unit=UNIT)
    dassert(rq, lambda r: r.function_code < 0x80)  # test for no error
    dassert(rr, _print)  # test the expected value

    rq = client.write_register(1, 10, unit=UNIT)
    rr = client.read_holding_registers(1, 1, unit=UNIT)
    dassert(rq, lambda r: r.function_code < 0x80)  # test for no error
    dassert(rr, _print)  # test the expected value

    rq = client.write_registers(1, [10] * 8, unit=UNIT)
    rr = client.read_input_registers(1, 8, unit=UNIT)
    dassert(rq, lambda r: r.function_code < 0x80)  # test for no error
    dassert(rr, _print)  # test the expected value

    arguments = {
        "read_address": 1,
        "read_count": 8,
        "write_address": 1,
        "write_registers": [20] * 8,
    }
    rq = client.readwrite_registers(**arguments, unit=UNIT)
    rr = client.read_input_registers(1, 8, unit=UNIT)
    dassert(rq, lambda r: r.registers == [20] * 8)  # test the expected value
    dassert(rr, _print)  # test the expected value

    # -----------------------------------------------------------------------#
    # close the client at some time later
    # -----------------------------------------------------------------------#
    IOLoop.current().add_timeout(IOLoop.current().time() + 1, client.close)
    IOLoop.current().add_timeout(IOLoop.current().time() + 2, protocol.stop)


# ---------------------------------------------------------------------------#
# choose the client you want
# ---------------------------------------------------------------------------#
# make sure to start an implementation to hit against. For this
# you can use an existing device, the reference implementation in the tools
# directory, or start a pymodbus server.
# ---------------------------------------------------------------------------#


def err(*args, **kwargs):
    """Handle error."""
    txt = f"Err {args} {kwargs}"
    log.error(txt)


def callback(protocol, future):
    """Call Callback."""
    log.debug("Client connected")
    if exp := future.exception():
        return err(exp)

    client = future.result()
    return begin_asynchronous_test(client, protocol)


if __name__ == "__main__":
    # ----------------------------------------------------------------------- #
    # Create temporary serial ports using SOCAT

    # socat -d -d PTY,link=/tmp/ptyp0,raw,echo=0,ispeed=9600 PTY,
    # link=/tmp/ttyp0,raw,echo=0,ospeed=9600

    # Default framer is ModbusRtuFramer
    # ----------------------------------------------------------------------- #

    # Rtu
    (
        protocol,
        future,
    ) = AsyncModbusSerialClient(  # pylint: disable=unpacking-non-sequence
        schedulers.IO_LOOP,
        method="rtu",
        port="/tmp/ptyp0",  # nosec
        baudrate=9600,
        timeout=2,
    )

    # Asci
    # from pymodbus.transaction import ModbusAsciiFramer
    # protocol, future = AsyncModbusSerialClient(schedulers.IO_LOOP,
    #                                            method="ascii",
    #                                            port="/dev/ptyp0",
    #                                            framer=ModbusAsciiFramer,
    #                                            baudrate=9600,
    #                                            timeout=2)

    # Binary
    # from pymodbus.transaction import ModbusBinaryFramer
    # protocol, future = AsyncModbusSerialClient(schedulers.IO_LOOP,
    #                                            method="binary",
    #                                            port="/dev/ptyp0",
    #                                            framer=ModbusBinaryFramer,
    #                                            baudrate=9600,
    #                                            timeout=2)
    future.add_done_callback(functools.partial(callback, protocol))



#!/usr/bin/env python3
# pylint: skip-file
"""Pydoc sub-class for generating documentation for entire packages.

Taken from: http://pyopengl.sourceforge.net/pydoc/OpenGLContext.pydoc.pydoc2.html
Author: Mike Fletcher
"""
import inspect
import logging
import os
import pydoc
import shutil
import string
from string import join, split, strip
import sys

_log = logging.getLogger(__name__)


def classify_class_attrs(cls):
    """Return list of attribute-descriptor tuples.

    For each name in dir(cls), the return list contains a 4-tuple
    with these elements:

        0. The name (a string).

        1. The kind of attribute this is, one of these strings:
               "class method"    created via classmethod()
               "static method"   created via staticmethod()
               "property"        created via property()
               "method"          any other flavor of method
               "data"            not a method

        2. The class which defined this attribute (a class).

        3. The object as obtained directly from the defining class"s
           __dict__, not via getattr.  This is especially important for
           data attributes:  C.data is just a data object, but
           C.__dict__["data"] may be a data descriptor with additional
           info, like a __doc__ string.

    Note: This version is patched to work with Zope Interface-bearing objects
    """
    mro = inspect.getmro(cls)
    names = dir(cls)
    result = []
    for name in names:
        # Get the object associated with the name.
        # Getting an obj from the __dict__ sometimes reveals more than
        # using getattr.  Static and class methods are dramatic examples.
        if name in cls.__dict__:
            obj = cls.__dict__[name]
        else:
            try:
                obj = getattr(cls, name)
            except AttributeError:
                continue

        # Figure out where it was defined.
        homecls = getattr(obj, "__objclass__", None)
        if homecls is None:
            # search the dicts.
            for base in mro:
                if name in base.__dict__:
                    homecls = base
                    break

        # Get the object again, in order to get it from the defining
        # __dict__ instead of via getattr (if possible).
        if homecls is not None and name in homecls.__dict__:
            obj = homecls.__dict__[name]

        # Also get the object via getattr.
        obj_via_getattr = getattr(cls, name)

        # Classify the object.
        if isinstance(obj, staticmethod):
            kind = "static method"
        elif isinstance(obj, classmethod):
            kind = "class method"
        elif isinstance(obj, property):
            kind = "property"
        elif inspect.ismethod(obj_via_getattr) or inspect.ismethoddescriptor(
            obj_via_getattr
        ):
            kind = "method"
        else:
            kind = "data"

        result.append((name, kind, homecls, obj))

    return result


inspect.classify_class_attrs = classify_class_attrs


class DefaultFormatter(pydoc.HTMLDoc):
    """Default formatter."""

    def docmodule(  # noqa: C901
        self, object, name=None, mod=None, packageContext=None, *ignored
    ):  # noqa: C901
        """Produce HTML documentation for a module object."""
        my_name = object.__name__  # ignore the passed-in name
        parts = split(my_name, ".")
        links = []
        for i in range(len(parts) - 1):
            links.append(
                '<a href="%s.html"><font color="#ffffff">%s</font></a>'
                % (join(parts[: i + 1], "."), parts[i])
            )
        linkedname = join(links + parts[-1:], ".")
        head = "<big><big><strong>%s</strong></big></big>" % linkedname
        try:
            path = inspect.getabsfile(object)
            url = path
            if sys.platform == "win32":
                import nturl2path

                url = nturl2path.pathname2url(path)
            filelink = '<a href="file:%s">%s</a>' % (url, path)
        except TypeError:
            filelink = "(built-in)"
        info = []
        if hasattr(object, "__version__"):
            version = str(object.__version__)
            if version[:11] == "$" + "Revision: " and version[-1:] == "$":
                version = strip(version[11:-1])
            info.append("version %s" % self.escape(version))
        if hasattr(object, "__date__"):
            info.append(self.escape(str(object.__date__)))
        if info:
            head = head + " (%s)" % join(info, ", ")
        result = self.heading(
            head, "#ffffff", "#7799ee", '<a href=".">index</a><br>' + filelink
        )

        modules = inspect.getmembers(object, inspect.ismodule)

        classes, cdict = [], {}
        for key, value in inspect.getmembers(object, inspect.isclass):
            if (inspect.getmodule(value) or object) is object:
                classes.append((key, value))
                cdict[key] = cdict[value] = "#" + key
        for key, value in classes:
            for base in value.__bases__:
                key, modname = base.__name__, base.__module__
                module = sys.modules.get(modname)
                if modname != my_name and module and hasattr(module, key):
                    if getattr(module, key) is base:
                        if key not in cdict:
                            cdict[key] = cdict[base] = modname + ".html#" + key
        funcs, fdict = [], {}
        for key, value in inspect.getmembers(object, inspect.isroutine):
            if inspect.isbuiltin(value) or inspect.getmodule(value) is object:
                funcs.append((key, value))
                fdict[key] = "#-" + key
                if inspect.isfunction(value):
                    fdict[value] = fdict[key]
        data = []
        for key, value in inspect.getmembers(object, pydoc.isdata):
            if key not in ["__builtins__", "__doc__"]:
                data.append((key, value))

        doc = self.markup(pydoc.getdoc(object), self.preformat, fdict, cdict)
        doc = doc and "<tt>%s</tt>" % doc
        result = result + "<p>%s</p>\n" % doc

        packageContext.clean(classes, object)
        packageContext.clean(funcs, object)
        packageContext.clean(data, object)

        if hasattr(object, "__path__"):
            modpkgs = []
            modnames = []
            for file in os.listdir(object.__path__[0]):
                path = os.path.join(object.__path__[0], file)
                modname = inspect.getmodulename(file)
                if modname and modname not in modnames:
                    modpkgs.append((modname, my_name, 0, 0))
                    modnames.append(modname)
                elif pydoc.ispackage(path):
                    modpkgs.append((file, my_name, 1, 0))
            modpkgs.sort()
            contents = self.multicolumn(modpkgs, self.modpkglink)
            # #   result = result + self.bigsection(
            # #   "Package Contents", "#ffffff", "#aa55cc", contents)
            result = result + self.moduleSection(object, packageContext)
        elif modules:
            # FIX   contents = self.multicolumn(
            # FIX   modules, lambda (key, value), s=self: s.modulelink(value))
            result = result + self.bigsection("Modules", "#fffff", "#aa55cc", contents)

        if classes:
            # FIX  classlist = map(lambda (key, value): value, classes)
            contents = [
                self.formattree(
                    inspect.getclasstree(classlist, 1), my_name  # noqa: F821
                )
            ]
            for key, value in classes:
                contents.append(self.document(value, key, my_name, fdict, cdict))
            result = result + self.bigsection(
                "Classes", "#ffffff", "#ee77aa", join(contents)
            )
        if funcs:
            contents = []
            for key, value in funcs:
                contents.append(self.document(value, key, my_name, fdict, cdict))
            result = result + self.bigsection(
                "Functions", "#ffffff", "#eeaa77", join(contents)
            )
        if data:
            contents = []
            for key, value in data:
                try:
                    contents.append(self.document(value, key))
                except Exception:  # nosec
                    pass
            result = result + self.bigsection(
                "Data", "#ffffff", "#55aa55", join(contents, "<br>\n")
            )
        if hasattr(object, "__author__"):
            contents = self.markup(str(object.__author__), self.preformat)
            result = result + self.bigsection("Author", "#ffffff", "#7799ee", contents)
        if hasattr(object, "__credits__"):
            contents = self.markup(str(object.__credits__), self.preformat)
            result = result + self.bigsection("Credits", "#ffffff", "#7799ee", contents)

        return result

    def classlink(self, object, modname):
        """Make a link for a class."""
        name, module = object.__name__, sys.modules.get(object.__module__)
        if hasattr(module, name) and getattr(module, name) is object:
            return '<a href="%s.html#%s">%s</a>' % (module.__name__, name, name)
        return pydoc.classname(object, modname)

    def moduleSection(self, object, packageContext):
        """Create a module-links section for the given object (module)"""
        modules = inspect.getmembers(object, inspect.ismodule)
        packageContext.clean(modules, object)
        packageContext.recurseScan(modules)

        if hasattr(object, "__path__"):
            modpkgs = []
            modnames = []
            for file in os.listdir(object.__path__[0]):
                path = os.path.join(object.__path__[0], file)
                modname = inspect.getmodulename(file)
                if modname and modname not in modnames:
                    modpkgs.append((modname, object.__name__, 0, 0))
                    modnames.append(modname)
                elif pydoc.ispackage(path):
                    modpkgs.append((file, object.__name__, 1, 0))
            modpkgs.sort()
            # do more recursion here...
            for (modname, name, ya, yo) in modpkgs:
                packageContext.addInteresting(join((object.__name__, modname), "."))
            items = []
            for (modname, name, ispackage, isshadowed) in modpkgs:
                try:
                    module = pydoc.safeimport("%s.%s" % (name, modname))
                    description, documentation = pydoc.splitdoc(inspect.getdoc(module))
                    if description:
                        txt = f"{self.modpkglink((modname, name, ispackage, isshadowed))} -- {description}"
                        items.append(txt)
                    else:
                        items.append(
                            self.modpkglink((modname, name, ispackage, isshadowed))
                        )
                except:  # noqa: E722
                    items.append(
                        self.modpkglink((modname, name, ispackage, isshadowed))
                    )
            contents = string.join(items, "<br>")
            result = self.bigsection("Package Contents", "#ffffff", "#aa55cc", contents)
        elif modules:
            result = self.bigsection("Modules", "#fffff", "#aa55cc", contents)
        else:
            result = ""
        return result


class AlreadyDone(Exception):
    """Already done."""

    pass


class PackageDocumentationGenerator:
    """A package document generator creates documentation.

    for an entire package using pydoc"s machinery.

    baseModules -- modules which will be included
        and whose included and children modules will be
        considered fair game for documentation
    destinationDirectory -- the directory into which
        the HTML documentation will be written
    recursion -- whether to add modules which are
        referenced by and/or children of base modules
    exclusions -- a list of modules whose contents will
        not be shown in any other module, commonly
        such modules as OpenGL.GL, wxPython.wx etc.
    recursionStops -- a list of modules which will
        explicitly stop recursion (i.e. they will never
        be included), even if they are children of base
        modules.
    formatter -- allows for passing in a custom formatter
        see DefaultFormatter for sample implementation.
    """

    def __init__(
        self,
        baseModules,
        destinationDirectory=".",
        recursion=1,
        exclusions=(),
        recursionStops=(),
        formatter=None,
    ):
        """Initialize."""
        self.destinationDirectory = os.path.abspath(destinationDirectory)
        self.exclusions = {}
        self.warnings = []
        self.baseSpecifiers = {}
        self.completed = {}
        self.recursionStops = {}
        self.recursion = recursion
        for stop in recursionStops:
            self.recursionStops[stop] = 1
        self.pending = []
        for exclusion in exclusions:
            try:
                self.exclusions[exclusion] = pydoc.locate(exclusion)
            except pydoc.ErrorDuringImport:
                self.warn(
                    """Unable to import the module %s which was specified as an exclusion module"""
                    % (repr(exclusion))
                )
        self.formatter = formatter or DefaultFormatter()
        for base in baseModules:
            self.addBase(base)

    def warn(self, message):
        """Get warnings are used for recoverable, but not necessarily ignorable conditions."""
        self.warnings.append(message)

    def info(self, message):
        """Get information/status report."""
        _log.debug(message)

    def addBase(self, specifier):
        """Set the base of the documentation set, only children of these modules will be documented."""
        try:
            self.baseSpecifiers[specifier] = pydoc.locate(specifier)
            self.pending.append(specifier)
        except pydoc.ErrorDuringImport:
            self.warn(
                """Unable to import the module %s which was specified as a base module"""
                % (repr(specifier))
            )

    def addInteresting(self, specifier):
        """Add a module to the list of interesting modules."""
        if self.checkScope(specifier):
            self.pending.append(specifier)
        else:
            self.completed[specifier] = 1

    def checkScope(self, specifier):
        """Check that the specifier is "in scope" for the recursion."""
        if not self.recursion:
            return 0
        items = string.split(specifier, ".")
        stopCheck = items[:]
        while stopCheck:
            name = string.join(items, ".")
            if self.recursionStops.get(name):
                return 0
            elif self.completed.get(name):
                return 0
            del stopCheck[-1]
        while items:
            if self.baseSpecifiers.get(string.join(items, ".")):
                return 1
            del items[-1]
        # was not within any given scope
        return 0

    def process(self):
        """Process.

        Having added all of the base and/or interesting modules,
        proceed to generate the appropriate documentation for each
        module in the appropriate directory, doing the recursion
        as we go.
        """
        try:
            while self.pending:
                try:
                    if self.pending[0] in self.completed:
                        raise AlreadyDone(self.pending[0])
                    self.info("""Start %s""" % (repr(self.pending[0])))
                    object = pydoc.locate(self.pending[0])
                    self.info("""   ... found %s""" % (repr(object.__name__)))
                except AlreadyDone:
                    pass
                except pydoc.ErrorDuringImport as value:
                    self.info("""   ... FAILED %s""" % (repr(value)))
                    self.warn(
                        """Unable to import the module %s""" % (repr(self.pending[0]))
                    )
                except (SystemError, SystemExit) as value:
                    self.info("""   ... FAILED %s""" % (repr(value)))
                    self.warn(
                        """Unable to import the module %s""" % (repr(self.pending[0]))
                    )
                except Exception as value:
                    self.info("""   ... FAILED %s""" % (repr(value)))
                    self.warn(
                        """Unable to import the module %s""" % (repr(self.pending[0]))
                    )
                else:
                    page = self.formatter.page(
                        pydoc.describe(object),
                        self.formatter.docmodule(
                            object,
                            object.__name__,
                            packageContext=self,
                        ),
                    )
                    file = open(
                        os.path.join(
                            self.destinationDirectory,
                            self.pending[0] + ".html",
                        ),
                        "w",
                    )
                    file.write(page)
                    file.close()
                    self.completed[self.pending[0]] = object
                del self.pending[0]
        finally:
            for item in self.warnings:
                _log.info(item)

    def clean(self, objectList, object):
        """Clean.

        Callback from the formatter object asking us to remove
        those items in the key, value pairs where the object is
        imported from one of the excluded modules.
        """
        for key, value in objectList[:]:
            for excludeObject in self.exclusions.values():
                if hasattr(excludeObject, key) and excludeObject is not object:
                    if getattr(excludeObject, key) is value or (
                        hasattr(excludeObject, "__name__")
                        and excludeObject.__name__ == "Numeric"
                    ):
                        objectList[:] = [(k, o) for k, o in objectList if k != key]

    def recurseScan(self, objectList):
        """Process the list of modules.

        trying to add each to the
        list of interesting modules.
        """
        for key, value in objectList:
            self.addInteresting(value.__name__)


# ---------------------------------------------------------------------------#
#  Main Runner
# ---------------------------------------------------------------------------#
if __name__ == "__main__":
    if not os.path.exists("./html"):
        os.mkdir("./html")

    print("Building Pydoc API Documentation")
    PackageDocumentationGenerator(
        baseModules=["pymodbus", "__builtin__"],
        destinationDirectory="./html/",
        exclusions=["math", "string"],
        recursionStops=[],
    ).process()

    if os.path.exists("../../../build"):
        shutil.move("html", "../../../build/pydoc")



"""Constants For Modbus Server/Client.

This is the single location for storing default
values for the servers and clients.
"""
import logging

from pymodbus.interfaces import Singleton


# set logging format and default level for library.
logging.basicConfig(
    format="%(asctime)s %(levelname)-5s %(module)s:%(lineno)s %(message)s",
    datefmt="%H:%M:%S",
    level=logging.WARNING,
)


class Defaults(Singleton):  # pylint: disable=too-few-public-methods
    """A collection of modbus default values.

    .. attribute:: Port

       The default modbus tcp server port (502)

    .. attribute:: TLSPort

       The default modbus tcp over tls server port (802)


    .. attribute:: Backoff

       The default exponential backoff delay (0.3 seconds)

    .. attribute:: Retries

       The default number of times a client should retry the given
       request before failing (3)

    .. attribute:: RetryOnEmpty

       A flag indicating if a transaction should be retried in the
       case that an empty response is received. This is useful for
       slow clients that may need more time to process a request.

    .. attribute:: RetryOnInvalid

       A flag indicating if a transaction should be retried in the
       case that an invalid response is received.

    .. attribute:: Timeout

       The default amount of time a client should wait for a request
       to be processed (3 seconds)

    .. attribute:: Reconnects

       The default number of times a client should attempt to reconnect
       before deciding the server is down (0)

    .. attribute:: TransactionId

       The starting transaction identifier number (0)

    .. attribute:: ProtocolId

       The modbus protocol id.  Currently this is set to 0 in all
       but proprietary implementations.

    .. attribute:: Slave

       The modbus slave address.  Currently this is set to 0x00 which
       means this request should be broadcast to all the slave devices
       (really means that all the devices should respond).

    .. attribute:: Baudrate

       The speed at which the data is transmitted over the serial line.
       This defaults to 19200.

    .. attribute:: Parity

       The type of checksum to use to verify data integrity. This can be
       on of the following::

         - (E)ven - 1 0 1 0 | P(0)
         - (O)dd  - 1 0 1 0 | P(1)
         - (N)one - 1 0 1 0 | no parity

       This defaults to (N)one.

    .. attribute:: Bytesize

       The number of bits in a byte of serial data.  This can be one of
       5, 6, 7, or 8. This defaults to 8.

    .. attribute:: Stopbits

       The number of bits sent after each character in a message to
       indicate the end of the byte.  This defaults to 1.

    .. attribute:: ZeroMode

       Indicates if the slave datastore should use indexing at 0 or 1.
       More about this can be read in section 4.4 of the modbus specification.

    .. attribute:: IgnoreMissingSlaves

       In case a request is made to a missing slave, this defines if an error
       should be returned or simply ignored. This is useful for the case of a
       serial server emulater where a request to a non-existent slave on a bus
       will never respond. The client in this case will simply timeout.

    .. attribute:: broadcast_enable

      When False unit_id 0 will be treated as any other unit_id. When True and
      the unit_id is 0 the server will execute all requests on all server
      contexts and not respond and the client will skip trying to receive a
      response. Default value False does not conform to Modbus spec but maintains
      legacy behavior for existing pymodbus users.

    """

    TcpPort = 502
    TlsPort = 802
    UdpPort = 502
    Backoff = 0.3
    CloseCommOnError = False
    HandleLocalEcho = False
    Retries = 3
    RetryOnEmpty = False
    RetryOnInvalid = False
    Timeout = 3
    Reconnects = 0
    TransactionId = 0
    Strict = True
    ProtocolId = 0
    Slave = 0x00
    Baudrate = 19200
    Parity = "N"
    Bytesize = 8
    Stopbits = 1
    ZeroMode = False
    IgnoreMissingSlaves = False
    ReadSize = 1024
    BroadcastEnable = False
    ReconnectDelay = 1000 * 60 * 5
    Count = 1


class ModbusStatus(Singleton):  # pylint: disable=too-few-public-methods
    """These represent various status codes in the modbus protocol.

    .. attribute:: Waiting

       This indicates that a modbus device is currently
       waiting for a given request to finish some running task.

    .. attribute:: Ready

       This indicates that a modbus device is currently
       free to perform the next request task.

    .. attribute:: On

       This indicates that the given modbus entity is on

    .. attribute:: Off

       This indicates that the given modbus entity is off

    .. attribute:: SlaveOn

       This indicates that the given modbus slave is running

    .. attribute:: SlaveOff

       This indicates that the given modbus slave is not running
    """

    Waiting = 0xFFFF
    Ready = 0x0000
    On = 0xFF00  # pylint: disable=invalid-name
    Off = 0x0000
    SlaveOn = 0xFF
    SlaveOff = 0x00


class Endian(Singleton):  # pylint: disable=too-few-public-methods
    """An enumeration representing the various byte endianness.

    .. attribute:: Auto

       This indicates that the byte order is chosen by the
       current native environment.

    .. attribute:: Big

       This indicates that the bytes are in little endian format

    .. attribute:: Little

       This indicates that the bytes are in big endian format

    .. note:: I am simply borrowing the format strings from the
       python struct module for my convenience.
    """

    Auto = "@"
    Big = ">"
    Little = "<"


class ModbusPlusOperation(Singleton):  # pylint: disable=too-few-public-methods
    """Represents the type of modbus plus request.

    .. attribute:: GetStatistics

       Operation requesting that the current modbus plus statistics
       be returned in the response.

    .. attribute:: ClearStatistics

       Operation requesting that the current modbus plus statistics
       be cleared and not returned in the response.
    """

    GetStatistics = 0x0003
    ClearStatistics = 0x0004


class DeviceInformation(Singleton):  # pylint: disable=too-few-public-methods
    """Represents what type of device information to read.

    .. attribute:: Basic

       This is the basic (required) device information to be returned.
       This includes VendorName, ProductCode, and MajorMinorRevision
       code.

    .. attribute:: Regular

       In addition to basic data objects, the device provides additional
       and optional identification and description data objects. All of
       the objects of this category are defined in the standard but their
       implementation is optional.

    .. attribute:: Extended

       In addition to regular data objects, the device provides additional
       and optional identification and description private data about the
       physical device itself. All of these data are device dependent.

    .. attribute:: Specific

       Request to return a single data object.
    """

    Basic = 0x01
    Regular = 0x02
    Extended = 0x03
    Specific = 0x04


class MoreData(Singleton):  # pylint: disable=too-few-public-methods
    """Represents the more follows condition.

    .. attribute:: Nothing

       This indicates that no more objects are going to be returned.

    .. attribute:: KeepReading

       This indicates that there are more objects to be returned.
    """

    Nothing = 0x00
    KeepReading = 0xFF


# ---------------------------------------------------------------------------#
#  Exported Identifiers
# ---------------------------------------------------------------------------#
__all__ = [
    "Defaults",
    "ModbusStatus",
    "Endian",
    "ModbusPlusOperation",
    "DeviceInformation",
    "MoreData",
]



"""Register Reading Request/Response."""
# pylint: disable=missing-type-doc
import struct

from pymodbus.constants import Defaults
from pymodbus.pdu import (
    ModbusExceptions as merror,
    ModbusRequest,
    ModbusResponse,
)


class ReadRegistersRequestBase(ModbusRequest):
    """Base class for reading a modbus register."""

    _rtu_frame_size = 8

    def __init__(self, address, count, unit=Defaults.Slave, **kwargs):
        """Initialize a new instance.

        :param address: The address to start the read from
        :param count: The number of registers to read
        :param unit: Modbus slave unit ID
        """
        super().__init__(unit, **kwargs)
        self.address = address
        self.count = count

    def encode(self):
        """Encode the request packet.

        :return: The encoded packet
        """
        return struct.pack(">HH", self.address, self.count)

    def decode(self, data):
        """Decode a register request packet.

        :param data: The request to decode
        """
        self.address, self.count = struct.unpack(">HH", data)

    def get_response_pdu_size(self):
        """Get response pdu size.

        Func_code (1 byte) + Byte Count(1 byte) + 2 * Quantity of Coils (n Bytes).
        """
        return 1 + 1 + 2 * self.count

    def __str__(self):
        """Return a string representation of the instance.

        :returns: A string representation of the instance
        """
        return f"ReadRegisterRequest ({self.address},{self.count})"


class ReadRegistersResponseBase(ModbusResponse):
    """Base class for responding to a modbus register read.

    The requested registers can be found in the .registers list.
    """

    _rtu_byte_count_pos = 2

    def __init__(self, values, unit=Defaults.Slave, **kwargs):
        """Initialize a new instance.

        :param values: The values to write to
        :param unit: Modbus slave unit ID
        """
        super().__init__(unit, **kwargs)

        #: A list of register values
        self.registers = values or []

    def encode(self):
        """Encode the response packet.

        :returns: The encoded packet
        """
        result = struct.pack(">B", len(self.registers) * 2)
        for register in self.registers:
            result += struct.pack(">H", register)
        return result

    def decode(self, data):
        """Decode a register response packet.

        :param data: The request to decode
        """
        byte_count = int(data[0])
        self.registers = []
        for i in range(1, byte_count + 1, 2):
            self.registers.append(struct.unpack(">H", data[i : i + 2])[0])

    def getRegister(self, index):  # pylint: disable=(invalid-name
        """Get the requested register.

        :param index: The indexed register to retrieve
        :returns: The request register
        """
        return self.registers[index]

    def __str__(self):
        """Return a string representation of the instance.

        :returns: A string representation of the instance
        """
        return f"{self.__class__.__name__} ({len(self.registers)})"


class ReadHoldingRegistersRequest(ReadRegistersRequestBase):
    """Read holding registers.

    This function code is used to read the contents of a contiguous block
    of holding registers in a remote device. The Request PDU specifies the
    starting register address and the number of registers. In the PDU
    Registers are addressed starting at zero. Therefore registers numbered
    1-16 are addressed as 0-15.
    """

    function_code = 3

    def __init__(self, address=None, count=None, unit=Defaults.Slave, **kwargs):
        """Initialize a new instance of the request.

        :param address: The starting address to read from
        :param count: The number of registers to read from address
        :param unit: Modbus slave unit ID
        """
        super().__init__(address, count, unit, **kwargs)

    def execute(self, context):
        """Run a read holding request against a datastore.

        :param context: The datastore to request from
        :returns: An initialized :py:class:`~pymodbus.register_read_message.ReadHoldingRegistersResponse`, or an :py:class:`~pymodbus.pdu.ExceptionResponse` if an error occurred
        """
        if not (1 <= self.count <= 0x7D):
            return self.doException(merror.IllegalValue)
        if not context.validate(self.function_code, self.address, self.count):
            return self.doException(merror.IllegalAddress)
        values = context.getValues(self.function_code, self.address, self.count)
        return ReadHoldingRegistersResponse(values)


class ReadHoldingRegistersResponse(ReadRegistersResponseBase):
    """Read holding registers.

    This function code is used to read the contents of a contiguous block
    of holding registers in a remote device. The Request PDU specifies the
    starting register address and the number of registers. In the PDU
    Registers are addressed starting at zero. Therefore registers numbered
    1-16 are addressed as 0-15.

    The requested registers can be found in the .registers list.
    """

    function_code = 3

    def __init__(self, values=None, **kwargs):
        """Initialize a new response instance.

        :param values: The resulting register values
        """
        super().__init__(values, **kwargs)


class ReadInputRegistersRequest(ReadRegistersRequestBase):
    """Read input registers.

    This function code is used to read from 1 to approx. 125 contiguous
    input registers in a remote device. The Request PDU specifies the
    starting register address and the number of registers. In the PDU
    Registers are addressed starting at zero. Therefore input registers
    numbered 1-16 are addressed as 0-15.
    """

    function_code = 4

    def __init__(self, address=None, count=None, unit=Defaults.Slave, **kwargs):
        """Initialize a new instance of the request.

        :param address: The starting address to read from
        :param count: The number of registers to read from address
        :param unit: Modbus slave unit ID
        """
        super().__init__(address, count, unit, **kwargs)

    def execute(self, context):
        """Run a read input request against a datastore.

        :param context: The datastore to request from
        :returns: An initialized :py:class:`~pymodbus.register_read_message.ReadInputRegistersResponse`, or an :py:class:`~pymodbus.pdu.ExceptionResponse` if an error occurred
        """
        if not (1 <= self.count <= 0x7D):
            return self.doException(merror.IllegalValue)
        if not context.validate(self.function_code, self.address, self.count):
            return self.doException(merror.IllegalAddress)
        values = context.getValues(self.function_code, self.address, self.count)
        return ReadInputRegistersResponse(values)


class ReadInputRegistersResponse(ReadRegistersResponseBase):
    """Read/write input registers.

    This function code is used to read from 1 to approx. 125 contiguous
    input registers in a remote device. The Request PDU specifies the
    starting register address and the number of registers. In the PDU
    Registers are addressed starting at zero. Therefore input registers
    numbered 1-16 are addressed as 0-15.

    The requested registers can be found in the .registers list.
    """

    function_code = 4

    def __init__(self, values=None, **kwargs):
        """Initialize a new response instance.

        :param values: The resulting register values
        """
        super().__init__(values, **kwargs)


class ReadWriteMultipleRegistersRequest(ModbusRequest):
    """Read/write multiple registers.

    This function code performs a combination of one read operation and one
    write operation in a single MODBUS transaction. The write
    operation is performed before the read.

    Holding registers are addressed starting at zero. Therefore holding
    registers 1-16 are addressed in the PDU as 0-15.

    The request specifies the starting address and number of holding
    registers to be read as well as the starting address, number of holding
    registers, and the data to be written. The byte count specifies the
    number of bytes to follow in the write data field."
    """

    function_code = 23
    _rtu_byte_count_pos = 10

    def __init__(self, **kwargs):
        """Initialize a new request message.

        :param read_address: The address to start reading from
        :param read_count: The number of registers to read from address
        :param write_address: The address to start writing to
        :param write_registers: The registers to write to the specified address
        """
        super().__init__(**kwargs)
        self.read_address = kwargs.get("read_address", 0x00)
        self.read_count = kwargs.get("read_count", 0)
        self.write_address = kwargs.get("write_address", 0x00)
        self.write_registers = kwargs.get("write_registers", None)
        if not hasattr(self.write_registers, "__iter__"):
            self.write_registers = [self.write_registers]
        self.write_count = len(self.write_registers)
        self.write_byte_count = self.write_count * 2

    def encode(self):
        """Encode the request packet.

        :returns: The encoded packet
        """
        result = struct.pack(
            ">HHHHB",
            self.read_address,
            self.read_count,
            self.write_address,
            self.write_count,
            self.write_byte_count,
        )
        for register in self.write_registers:
            result += struct.pack(">H", register)
        return result

    def decode(self, data):
        """Decode the register request packet.

        :param data: The request to decode
        """
        (
            self.read_address,
            self.read_count,
            self.write_address,
            self.write_count,
            self.write_byte_count,
        ) = struct.unpack(">HHHHB", data[:9])
        self.write_registers = []
        for i in range(9, self.write_byte_count + 9, 2):
            register = struct.unpack(">H", data[i : i + 2])[0]
            self.write_registers.append(register)

    def execute(self, context):
        """Run a write single register request against a datastore.

        :param context: The datastore to request from
        :returns: An initialized :py:class:`~pymodbus.register_read_message.ReadWriteMultipleRegistersResponse`, or an :py:class:`~pymodbus.pdu.ExceptionResponse` if an error occurred
        """
        if not (1 <= self.read_count <= 0x07D):
            return self.doException(merror.IllegalValue)
        if not 1 <= self.write_count <= 0x079:
            return self.doException(merror.IllegalValue)
        if self.write_byte_count != self.write_count * 2:
            return self.doException(merror.IllegalValue)
        if not context.validate(
            self.function_code, self.write_address, self.write_count
        ):
            return self.doException(merror.IllegalAddress)
        if not context.validate(self.function_code, self.read_address, self.read_count):
            return self.doException(merror.IllegalAddress)
        context.setValues(self.function_code, self.write_address, self.write_registers)
        registers = context.getValues(
            self.function_code, self.read_address, self.read_count
        )
        return ReadWriteMultipleRegistersResponse(registers)

    def get_response_pdu_size(self):
        """Get response pdu size.

        Func_code (1 byte) + Byte Count(1 byte) + 2 * Quantity of Coils (n Bytes)
        :return:
        """
        return 1 + 1 + 2 * self.read_count

    def __str__(self):
        """Return a string representation of the instance.

        :returns: A string representation of the instance
        """
        params = (
            self.read_address,
            self.read_count,
            self.write_address,
            self.write_count,
        )
        return (
            "ReadWriteNRegisterRequest R(%d,%d) W(%d,%d)"  # pylint: disable=consider-using-f-string
            % params
        )


class ReadWriteMultipleRegistersResponse(ModbusResponse):
    """Read/write multiple registers.

    The normal response contains the data from the group of registers that
    were read. The byte count field specifies the quantity of bytes to
    follow in the read data field.

    The requested registers can be found in the .registers list.
    """

    function_code = 23
    _rtu_byte_count_pos = 2

    def __init__(self, values=None, **kwargs):
        """Initialize a new instance.

        :param values: The register values to write
        """
        super().__init__(**kwargs)
        self.registers = values or []

    def encode(self):
        """Encode the response packet.

        :returns: The encoded packet
        """
        result = struct.pack(">B", len(self.registers) * 2)
        for register in self.registers:
            result += struct.pack(">H", register)
        return result

    def decode(self, data):
        """Decode the register response packet.

        :param data: The response to decode
        """
        bytecount = int(data[0])
        for i in range(1, bytecount, 2):
            self.registers.append(struct.unpack(">H", data[i : i + 2])[0])

    def __str__(self):
        """Return a string representation of the instance.

        :returns: A string representation of the instance
        """
        return f"ReadWriteNRegisterResponse ({len(self.registers)})"


# ---------------------------------------------------------------------------#
#  Exported symbols
# ---------------------------------------------------------------------------#
__all__ = [
    "ReadHoldingRegistersRequest",
    "ReadHoldingRegistersResponse",
    "ReadInputRegistersRequest",
    "ReadInputRegistersResponse",
    "ReadRegistersResponseBase",
    "ReadWriteMultipleRegistersRequest",
    "ReadWriteMultipleRegistersResponse",
]



"""These are a collection of helper methods.

that can be used to save a modbus server context to file for backup,
checkpointing, or any other purpose. There use is very
simple::

    context = server.context
    saver   = JsonDatastoreSaver(context)
    saver.save()

These can then be re-opened by the parsers in the
modbus_mapping module. At the moment, the supported
output formats are:

* csv
* json
* xml

To implement your own, simply subclass ModbusDatastoreSaver
and supply the needed callbacks for your given format:

* handle_store_start(self, store)
* handle_store_end(self, store)
* handle_slave_start(self, slave)
* handle_slave_end(self, slave)
* handle_save_start(self)
* handle_save_end(self)
"""
import json
import xml.etree.ElementTree as xml  # nosec


class ModbusDatastoreSaver:
    """An abstract base class.

    that can be used to implement
    a persistence format for the modbus server context. In
    order to use it, just complete the necessary callbacks
    (SAX style) that your persistence format needs.
    """

    def __init__(self, context, path=None):
        """Initialize a new instance of the saver.

        :param context: The modbus server context
        :param path: The output path to save to
        """
        self.context = context
        self.path = path or "modbus-context-dump"

    def save(self):
        """Save the context to file.

        which calls the various callbacks which the sub classes will implement.
        """
        with open(  # pylint: disable=unspecified-encoding
            self.path, "w"
        ) as self.file_handle:  # pylint: disable=attribute-defined-outside-init
            self.handle_save_start()
            for slave_name, slave in self.context:
                self.handle_slave_start(slave_name)
                for store_name, store in slave.store.iteritems():
                    self.handle_store_start(store_name)
                    self.handle_store_values(iter(store))  # pylint: disable=no-member
                    self.handle_store_end(store_name)
                self.handle_slave_end(slave_name)
            self.handle_save_end()

    # ------------------------------------------------------------
    # predefined state machine callbacks
    # ------------------------------------------------------------
    def handle_save_start(self):
        """Handle save start."""

    def handle_store_start(self, store):
        """Handle store start."""

    def handle_store_end(self, store):
        """Handle store end."""

    def handle_slave_start(self, slave):
        """Handle slave start."""

    def handle_slave_end(self, slave):
        """Handle slave end."""

    def handle_save_end(self):
        """Handle save end."""


# ---------------------------------------------------------------- #
# Implementations of the data store savers
# ---------------------------------------------------------------- #
class JsonDatastoreSaver(ModbusDatastoreSaver):
    """An implementation of the modbus datastore saver.

    that persists the context as a json document.
    """

    _context = None
    _store = None
    _slave = None

    STORE_NAMES = {
        "i": "input-registers",
        "d": "discretes",
        "h": "holding-registers",
        "c": "coils",
    }

    def handle_save_start(self):
        """Handle save start."""
        self._context = {}

    def handle_slave_start(self, slave):
        """Handle slave start."""
        self._context[hex(slave)] = self._slave = {}

    def handle_store_start(self, store):
        """Handle store start."""
        self._store = self.STORE_NAMES[store]

    def handle_store_values(self, values):
        """Handle store values."""
        self._slave[self._store] = dict(values)

    def handle_save_end(self):
        """Handle save end."""
        json.dump(self._context, self.file_handle)


class CsvDatastoreSaver(ModbusDatastoreSaver):
    """An implementation of the modbus datastore saver.

    that persists the context as a csv document.
    """

    _context = None
    _store = None
    _line = None
    NEWLINE = "\r\n"
    HEADER = "slave,store,address,value" + NEWLINE
    STORE_NAMES = {
        "i": "i",
        "d": "d",
        "h": "h",
        "c": "c",
    }

    def handle_save_start(self):
        """Handle save start."""
        self.file_handle.write(self.HEADER)

    def handle_slave_start(self, slave):
        """Handle slave start."""
        self._line = [str(slave)]

    def handle_store_start(self, store):
        """Handle store start."""
        self._line.append(self.STORE_NAMES[store])

    def handle_store_values(self, values):
        """Handle store values."""
        self.file_handle.writelines(self.handle_store_value(values))

    def handle_store_end(self, store):
        """Handle store end."""
        self._line.pop()

    def handle_store_value(self, values):
        """Handle store value."""
        for val_a, val_v in values:
            yield ",".join(self._line + [str(val_a), str(val_v)]) + self.NEWLINE


class XmlDatastoreSaver(ModbusDatastoreSaver):
    """An implementation of the modbus datastore saver.

    that persists the context as a XML document.
    """

    _context = None
    _store = None

    STORE_NAMES = {
        "i": "input-registers",
        "d": "discretes",
        "h": "holding-registers",
        "c": "coils",
    }

    def handle_save_start(self):
        """Handle save start."""
        self._context = xml.Element("context")
        self._root = xml.ElementTree(  # pylint: disable=attribute-defined-outside-init
            self._context
        )

    def handle_slave_start(self, slave):
        """Handle slave start."""
        self._slave = xml.SubElement(  # pylint: disable=attribute-defined-outside-init
            self._context, "slave"
        )
        self._slave.set("id", str(slave))

    def handle_store_start(self, store):
        """Handle store start."""
        self._store = xml.SubElement(self._slave, "store")
        self._store.set("function", self.STORE_NAMES[store])

    def handle_store_values(self, values):
        """Handle store values."""
        for address, value in values:
            entry = xml.SubElement(self._store, "entry")
            entry.text = str(value)
            entry.set("address", str(address))

    def handle_save_end(self):
        """Handle save end."""
        self._root.write(self.file_handle)



"""Test remote datastore."""
import unittest
from unittest import mock

from pymodbus.bit_read_message import ReadCoilsResponse
from pymodbus.bit_write_message import WriteMultipleCoilsResponse
from pymodbus.datastore.remote import RemoteSlaveContext
from pymodbus.exceptions import NotImplementedException
from pymodbus.pdu import ExceptionResponse
from pymodbus.register_read_message import ReadInputRegistersResponse


class RemoteModbusDataStoreTest(unittest.TestCase):
    """Unittest for the pymodbus.datastore.remote module."""

    def test_remote_slave_context(self):
        """Test a modbus remote slave context"""
        context = RemoteSlaveContext(None)
        self.assertNotEqual(str(context), None)
        self.assertRaises(
            NotImplementedException,
            lambda: context.reset(),  # pylint: disable=unnecessary-lambda
        )

    def test_remote_slave_set_values(self):
        """Test setting values against a remote slave context"""
        client = mock.MagicMock()
        client.write_coils = lambda a, b: WriteMultipleCoilsResponse()

        context = RemoteSlaveContext(client)
        context.setValues(1, 0, [1])

    def test_remote_slave_get_values(self):
        """Test getting values from a remote slave context"""
        client = mock.MagicMock()
        client.read_coils = lambda a, b: ReadCoilsResponse([1] * 10)
        client.read_input_registers = lambda a, b: ReadInputRegistersResponse([10] * 10)
        client.read_holding_registers = lambda a, b: ExceptionResponse(0x15)

        context = RemoteSlaveContext(client)
        context.validate(1, 0, 10)
        result = context.getValues(1, 0, 10)
        self.assertEqual(result, [1] * 10)

        context.validate(4, 0, 10)
        result = context.getValues(4, 0, 10)
        self.assertEqual(result, [10] * 10)

        context.validate(3, 0, 10)
        result = context.getValues(3, 0, 10)
        self.assertNotEqual(result, [10] * 10)

    def test_remote_slave_validate_values(self):
        """Test validating against a remote slave context"""
        client = mock.MagicMock()
        client.read_coils = lambda a, b: ReadCoilsResponse([1] * 10)
        client.read_input_registers = lambda a, b: ReadInputRegistersResponse([10] * 10)
        client.read_holding_registers = lambda a, b: ExceptionResponse(0x15)

        context = RemoteSlaveContext(client)
        result = context.validate(1, 0, 10)
        self.assertTrue(result)

        result = context.validate(4, 0, 10)
        self.assertTrue(result)

        result = context.validate(3, 0, 10)
        self.assertFalse(result)



#!/usr/bin/env python3
# pylint: disable=missing-type-doc
"""Pymodbus Performance Example.

The following is an quick performance check of the synchronous
modbus client.
"""
from concurrent.futures import ThreadPoolExecutor as eWorker, as_completed

# --------------------------------------------------------------------------- #
# import the necessary modules
# --------------------------------------------------------------------------- #
import logging
import os
from threading import Lock, Thread as tWorker
from time import time

from pymodbus.client import ModbusTcpClient


try:
    from multiprocessing import Process as mWorker, log_to_stderr
except ImportError:
    log_to_stderr = logging.getLogger

# --------------------------------------------------------------------------- #
# choose between threads or processes
# --------------------------------------------------------------------------- #

# from multiprocessing import Process as Worker
# from threading import Thread as Worker
_thread_lock = Lock()
# --------------------------------------------------------------------------- #
# initialize the test
# --------------------------------------------------------------------------- #
# Modify the parameters below to control how we are testing the client:
#
# * workers - the number of workers to use at once
# * cycles  - the total number of requests to send
# * host    - the host to send the requests to
# --------------------------------------------------------------------------- #
workers = 10  # pylint: disable=invalid-name
cycles = 1000  # pylint: disable=invalid-name
host = "127.0.0.1"  # pylint: disable=invalid-name


# --------------------------------------------------------------------------- #
# perform the test
# --------------------------------------------------------------------------- #
# This test is written such that it can be used by many threads of processes
# although it should be noted that there are performance penalties
# associated with each strategy.
# --------------------------------------------------------------------------- #
def single_client_test(n_host, n_cycles):
    """Perform a single threaded test of a synchronous client against the specified host

    :param n_host: The host to connect to
    :param n_cycles: The number of iterations to perform
    """
    logger = log_to_stderr()
    logger.setLevel(logging.WARNING)
    txt = f"starting worker: {os.getpid()}"
    logger.debug(txt)

    try:
        count = 0
        client = ModbusTcpClient(n_host, port=5020)
        while count < n_cycles:
            client.read_holding_registers(10, 123, slave=1)
            count += 1
    except Exception:  # pylint: disable=broad-except
        logger.exception("failed to run test successfully")
    txt = f"finished worker: {os.getpid()}"
    logger.debug(txt)


def multiprocessing_test(func, extras):
    """Multiprocessing test."""
    start_time = time()
    procs = [mWorker(target=func, args=extras) for _ in range(workers)]

    any(p.start() for p in procs)  # start the workers
    any(p.join() for p in procs)  # wait for the workers to finish
    return start_time


def thread_test(func, extras):
    """Thread test."""
    start_time = time()
    procs = [tWorker(target=func, args=extras) for _ in range(workers)]

    any(p.start() for p in procs)  # start the workers
    any(p.join() for p in procs)  # wait for the workers to finish
    return start_time


def thread_pool_exe_test(func, extras):
    """Thread pool exe."""
    start_time = time()
    with eWorker(max_workers=workers, thread_name_prefix="Perform") as exe:
        futures = {exe.submit(func, *extras): job for job in range(workers)}
        for future in as_completed(futures):
            future.result()
    return start_time


# --------------------------------------------------------------------------- #
# run our test and check results
# --------------------------------------------------------------------------- #
# We shard the total number of requests to perform between the number of
# threads that was specified. We then start all the threads and block on
# them to finish. This may need to switch to another mechanism to signal
# finished as the process/thread start up/shut down may skew the test a bit.

# RTU 32 requests/second @9600
# TCP 31430 requests/second

# --------------------------------------------------------------------------- #


if __name__ == "__main__":
    args = (host, int(cycles * 1.0 / workers))
    # with Worker(max_workers=workers, thread_name_prefix="Perform") as exe:
    #     futures = {exe.submit(single_client_test, *args): job for job in range(workers)}
    #     for future in as_completed(futures):
    #         data = future.result()
    # for _ in range(workers):
    #    futures.append(Worker.submit(single_client_test, args=args))
    # procs = [Worker(target=single_client_test, args=args)
    #          for _ in range(workers)]

    # any(p.start() for p in procs)   # start the workers
    # any(p.join() for p in procs)   # wait for the workers to finish
    # start = multiprocessing_test(single_client_test, args)
    # start = thread_pool_exe_test(single_client_test, args)
    for tester in (multiprocessing_test, thread_test, thread_pool_exe_test):
        print(tester.__name__)
        start = tester(single_client_test, args)
        stop = time()
        print(f"{(1.0 * cycles) / (stop - start)} requests/second")
        print(
            f"time taken to complete {cycles} cycle by "
            f"{workers} workers is {stop - start} seconds"
        )
        print()



"""Test other messages."""
import unittest
from unittest import mock

import pymodbus.other_message as pymodbus_message


class ModbusOtherMessageTest(unittest.TestCase):
    """Unittest for the pymodbus.other_message module."""

    def setUp(self):
        """Do setup."""
        self.requests = [
            pymodbus_message.ReadExceptionStatusRequest,
            pymodbus_message.GetCommEventCounterRequest,
            pymodbus_message.GetCommEventLogRequest,
            pymodbus_message.ReportSlaveIdRequest,
        ]

        self.responses = [
            lambda: pymodbus_message.ReadExceptionStatusResponse(0x12),
            lambda: pymodbus_message.GetCommEventCounterResponse(0x12),
            pymodbus_message.GetCommEventLogResponse,
            lambda: pymodbus_message.ReportSlaveIdResponse(0x12),
        ]

    def tearDown(self):
        """Clean up the test environment."""
        del self.requests
        del self.responses

    def test_other_messages_to_string(self):
        """Test other messages to string."""
        for message in self.requests:
            self.assertNotEqual(str(message()), None)
        for message in self.responses:
            self.assertNotEqual(str(message()), None)

    def test_read_exception_status(self):
        """Test read exception status."""
        request = pymodbus_message.ReadExceptionStatusRequest()
        request.decode(b"\x12")
        self.assertEqual(request.encode(), b"")
        self.assertEqual(request.execute().function_code, 0x07)

        response = pymodbus_message.ReadExceptionStatusResponse(0x12)
        self.assertEqual(response.encode(), b"\x12")
        response.decode(b"\x12")
        self.assertEqual(response.status, 0x12)

    def test_get_comm_event_counter(self):
        """Test get comm event counter."""
        request = pymodbus_message.GetCommEventCounterRequest()
        request.decode(b"\x12")
        self.assertEqual(request.encode(), b"")
        self.assertEqual(request.execute().function_code, 0x0B)

        response = pymodbus_message.GetCommEventCounterResponse(0x12)
        self.assertEqual(response.encode(), b"\x00\x00\x00\x12")
        response.decode(b"\x00\x00\x00\x12")
        self.assertEqual(response.status, True)
        self.assertEqual(response.count, 0x12)

        response.status = False
        self.assertEqual(response.encode(), b"\xFF\xFF\x00\x12")

    def test_get_comm_event_log(self):
        """Test get comm event log."""
        request = pymodbus_message.GetCommEventLogRequest()
        request.decode(b"\x12")
        self.assertEqual(request.encode(), b"")
        self.assertEqual(request.execute().function_code, 0x0C)

        response = pymodbus_message.GetCommEventLogResponse()
        self.assertEqual(response.encode(), b"\x06\x00\x00\x00\x00\x00\x00")
        response.decode(b"\x06\x00\x00\x00\x12\x00\x12")
        self.assertEqual(response.status, True)
        self.assertEqual(response.message_count, 0x12)
        self.assertEqual(response.event_count, 0x12)
        self.assertEqual(response.events, [])

        response.status = False
        self.assertEqual(response.encode(), b"\x06\xff\xff\x00\x12\x00\x12")

    def test_get_comm_event_log_with_events(self):
        """Test get comm event log with events."""
        response = pymodbus_message.GetCommEventLogResponse(events=[0x12, 0x34, 0x56])
        self.assertEqual(response.encode(), b"\x09\x00\x00\x00\x00\x00\x00\x12\x34\x56")
        response.decode(b"\x09\x00\x00\x00\x12\x00\x12\x12\x34\x56")
        self.assertEqual(response.status, True)
        self.assertEqual(response.message_count, 0x12)
        self.assertEqual(response.event_count, 0x12)
        self.assertEqual(response.events, [0x12, 0x34, 0x56])

    def test_report_slave_id_request(self):
        """Test report slave id request."""
        with mock.patch("pymodbus.other_message.DeviceInformationFactory") as dif:
            # First test regular identity strings
            identity = {
                0x00: "VN",  # VendorName
                0x01: "PC",  # ProductCode
                0x02: "REV",  # MajorMinorRevision
                0x03: "VU",  # VendorUrl
                0x04: "PN",  # ProductName
                0x05: "MN",  # ModelName
                0x06: "UAN",  # UserApplicationName
                0x07: "RA",  # reserved
                0x08: "RB",  # reserved
            }
            dif.get.return_value = identity
            expected_identity = "-".join(identity.values()).encode()

            request = pymodbus_message.ReportSlaveIdRequest()
            response = request.execute()
            self.assertEqual(response.identifier, expected_identity)

            # Change to byte strings and test again (final result should be the same)
            identity = {
                0x00: b"VN",  # VendorName
                0x01: b"PC",  # ProductCode
                0x02: b"REV",  # MajorMinorRevision
                0x03: b"VU",  # VendorUrl
                0x04: b"PN",  # ProductName
                0x05: b"MN",  # ModelName
                0x06: b"UAN",  # UserApplicationName
                0x07: b"RA",  # reserved
                0x08: b"RB",  # reserved
            }
            dif.get.return_value = identity

            request = pymodbus_message.ReportSlaveIdRequest()
            response = request.execute()
            self.assertEqual(response.identifier, expected_identity)

    def test_report_slave_id(self):
        """Test report slave id."""
        with mock.patch("pymodbus.other_message.DeviceInformationFactory") as dif:
            dif.get.return_value = {}
            request = pymodbus_message.ReportSlaveIdRequest()
            request.decode(b"\x12")
            self.assertEqual(request.encode(), b"")
            self.assertEqual(request.execute().function_code, 0x11)

            response = pymodbus_message.ReportSlaveIdResponse(
                request.execute().identifier, True
            )

            self.assertEqual(response.encode(), b"\tPymodbus\xff")
            response.decode(b"\x03\x12\x00")
            self.assertEqual(response.status, False)
            self.assertEqual(response.identifier, b"\x12\x00")

            response.status = False
            self.assertEqual(response.encode(), b"\x03\x12\x00\x00")



"""Payload Utilities Test Fixture.

This fixture tests the functionality of the payload
utilities.

* PayloadBuilder
* PayloadDecoder
"""
import unittest

from pymodbus.constants import Endian
from pymodbus.exceptions import ParameterException
from pymodbus.payload import BinaryPayloadBuilder, BinaryPayloadDecoder


# ---------------------------------------------------------------------------#
#  Fixture
# ---------------------------------------------------------------------------#


class ModbusPayloadUtilityTests(unittest.TestCase):
    """Modbus payload utility tests."""

    # ----------------------------------------------------------------------- #
    # Setup/TearDown
    # ----------------------------------------------------------------------- #

    def setUp(self):
        """Initialize the test environment and builds request/result encoding pairs."""
        self.little_endian_payload = (
            b"\x01\x02\x00\x03\x00\x00\x00\x04\x00\x00\x00\x00"
            b"\x00\x00\x00\xff\xfe\xff\xfd\xff\xff\xff\xfc\xff"
            b"\xff\xff\xff\xff\xff\xff\x00\x00\xa0\x3f\x00\x00"
            b"\x00\x00\x00\x00\x19\x40\x01\x00\x74\x65\x73\x74"
            b"\x11"
        )

        self.big_endian_payload = (
            b"\x01\x00\x02\x00\x00\x00\x03\x00\x00\x00\x00\x00"
            b"\x00\x00\x04\xff\xff\xfe\xff\xff\xff\xfd\xff\xff"
            b"\xff\xff\xff\xff\xff\xfc\x3f\xa0\x00\x00\x40\x19"
            b"\x00\x00\x00\x00\x00\x00\x00\x01\x74\x65\x73\x74"
            b"\x11"
        )

        self.bitstring = [True, False, False, False, True, False, False, False]

    def tearDown(self):
        """Clean up the test environment"""

    # ----------------------------------------------------------------------- #
    # Payload Builder Tests
    # ----------------------------------------------------------------------- #

    def test_little_endian_payload_builder(self):
        """Test basic bit message encoding/decoding"""
        builder = BinaryPayloadBuilder(byteorder=Endian.Little, wordorder=Endian.Little)
        builder.add_8bit_uint(1)
        builder.add_16bit_uint(2)
        builder.add_32bit_uint(3)
        builder.add_64bit_uint(4)
        builder.add_8bit_int(-1)
        builder.add_16bit_int(-2)
        builder.add_32bit_int(-3)
        builder.add_64bit_int(-4)
        builder.add_32bit_float(1.25)
        builder.add_64bit_float(6.25)
        builder.add_16bit_uint(1)  # placeholder
        builder.add_string(b"test")
        builder.add_bits(self.bitstring)
        self.assertEqual(self.little_endian_payload, builder.to_string())

    def test_big_endian_payload_builder(self):
        """Test basic bit message encoding/decoding"""
        builder = BinaryPayloadBuilder(byteorder=Endian.Big)
        builder.add_8bit_uint(1)
        builder.add_16bit_uint(2)
        builder.add_32bit_uint(3)
        builder.add_64bit_uint(4)
        builder.add_8bit_int(-1)
        builder.add_16bit_int(-2)
        builder.add_32bit_int(-3)
        builder.add_64bit_int(-4)
        builder.add_32bit_float(1.25)
        builder.add_64bit_float(6.25)
        builder.add_16bit_uint(1)  # placeholder
        builder.add_string("test")
        builder.add_bits(self.bitstring)
        self.assertEqual(self.big_endian_payload, builder.to_string())

    def test_payload_builder_reset(self):
        """Test basic bit message encoding/decoding"""
        builder = BinaryPayloadBuilder()
        builder.add_8bit_uint(0x12)
        builder.add_8bit_uint(0x34)
        builder.add_8bit_uint(0x56)
        builder.add_8bit_uint(0x78)
        self.assertEqual(b"\x12\x34\x56\x78", builder.to_string())
        self.assertEqual([b"\x12\x34", b"\x56\x78"], builder.build())
        builder.reset()
        self.assertEqual(b"", builder.to_string())
        self.assertEqual([], builder.build())

    def test_payload_builder_with_raw_payload(self):
        """Test basic bit message encoding/decoding"""
        _coils1 = [
            False,
            False,
            True,
            True,
            False,
            True,
            False,
            False,
            False,
            False,
            False,
            True,
            False,
            False,
            True,
            False,
            False,
            True,
            True,
            True,
            True,
            False,
            False,
            False,
            False,
            True,
            False,
            True,
            False,
            True,
            True,
            False,
        ]
        _coils2 = [
            False,
            False,
            False,
            True,
            False,
            False,
            True,
            False,
            False,
            False,
            True,
            True,
            False,
            True,
            False,
            False,
            False,
            True,
            False,
            True,
            False,
            True,
            True,
            False,
            False,
            True,
            True,
            True,
            True,
            False,
            False,
            False,
        ]

        builder = BinaryPayloadBuilder(
            [b"\x12", b"\x34", b"\x56", b"\x78"], repack=True
        )
        self.assertEqual(b"\x12\x34\x56\x78", builder.to_string())
        self.assertEqual([13330, 30806], builder.to_registers())
        coils = builder.to_coils()
        self.assertEqual(_coils1, coils)

        builder = BinaryPayloadBuilder(
            [b"\x12", b"\x34", b"\x56", b"\x78"], byteorder=Endian.Big
        )
        self.assertEqual(b"\x12\x34\x56\x78", builder.to_string())
        self.assertEqual([4660, 22136], builder.to_registers())
        self.assertEqual("\x12\x34\x56\x78", str(builder))
        coils = builder.to_coils()
        self.assertEqual(_coils2, coils)

    # ----------------------------------------------------------------------- #
    # Payload Decoder Tests
    # ----------------------------------------------------------------------- #

    def test_little_endian_payload_decoder(self):
        """Test basic bit message encoding/decoding"""
        decoder = BinaryPayloadDecoder(
            self.little_endian_payload, byteorder=Endian.Little, wordorder=Endian.Little
        )
        self.assertEqual(1, decoder.decode_8bit_uint())
        self.assertEqual(2, decoder.decode_16bit_uint())
        self.assertEqual(3, decoder.decode_32bit_uint())
        self.assertEqual(4, decoder.decode_64bit_uint())
        self.assertEqual(-1, decoder.decode_8bit_int())
        self.assertEqual(-2, decoder.decode_16bit_int())
        self.assertEqual(-3, decoder.decode_32bit_int())
        self.assertEqual(-4, decoder.decode_64bit_int())
        self.assertEqual(1.25, decoder.decode_32bit_float())
        self.assertEqual(6.25, decoder.decode_64bit_float())
        self.assertEqual(None, decoder.skip_bytes(2))
        self.assertEqual("test", decoder.decode_string(4).decode())
        self.assertEqual(self.bitstring, decoder.decode_bits())

    def test_big_endian_payload_decoder(self):
        """Test basic bit message encoding/decoding"""
        decoder = BinaryPayloadDecoder(self.big_endian_payload, byteorder=Endian.Big)
        self.assertEqual(1, decoder.decode_8bit_uint())
        self.assertEqual(2, decoder.decode_16bit_uint())
        self.assertEqual(3, decoder.decode_32bit_uint())
        self.assertEqual(4, decoder.decode_64bit_uint())
        self.assertEqual(-1, decoder.decode_8bit_int())
        self.assertEqual(-2, decoder.decode_16bit_int())
        self.assertEqual(-3, decoder.decode_32bit_int())
        self.assertEqual(-4, decoder.decode_64bit_int())
        self.assertEqual(1.25, decoder.decode_32bit_float())
        self.assertEqual(6.25, decoder.decode_64bit_float())
        self.assertEqual(None, decoder.skip_bytes(2))
        self.assertEqual(b"test", decoder.decode_string(4))
        self.assertEqual(self.bitstring, decoder.decode_bits())

    def test_payload_decoder_reset(self):
        """Test the payload decoder reset functionality"""
        decoder = BinaryPayloadDecoder(b"\x12\x34")
        self.assertEqual(0x12, decoder.decode_8bit_uint())
        self.assertEqual(0x34, decoder.decode_8bit_uint())
        decoder.reset()
        self.assertEqual(0x3412, decoder.decode_16bit_uint())

    def test_payload_decoder_register_factory(self):
        """Test the payload decoder reset functionality"""
        payload = [1, 2, 3, 4]
        decoder = BinaryPayloadDecoder.fromRegisters(payload, byteorder=Endian.Little)
        encoded = b"\x00\x01\x00\x02\x00\x03\x00\x04"
        self.assertEqual(encoded, decoder.decode_string(8))

        decoder = BinaryPayloadDecoder.fromRegisters(payload, byteorder=Endian.Big)
        encoded = b"\x00\x01\x00\x02\x00\x03\x00\x04"
        self.assertEqual(encoded, decoder.decode_string(8))

        self.assertRaises(
            ParameterException, lambda: BinaryPayloadDecoder.fromRegisters("abcd")
        )

    def test_payload_decoder_coil_factory(self):
        """Test the payload decoder reset functionality"""
        payload = [1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1]
        decoder = BinaryPayloadDecoder.fromCoils(payload, byteorder=Endian.Little)
        encoded = b"\x88\x11"
        self.assertEqual(encoded, decoder.decode_string(2))

        decoder = BinaryPayloadDecoder.fromCoils(payload, byteorder=Endian.Big)
        encoded = b"\x88\x11"
        self.assertEqual(encoded, decoder.decode_string(2))

        self.assertRaises(
            ParameterException, lambda: BinaryPayloadDecoder.fromCoils("abcd")
        )



#!/usr/bin/env python3
"""Pymodbus Asynchronous Client Examples.

The following is an example of how to use the asynchronous serial modbus
client implementation from pymodbus with twisted.
"""
import logging

from twisted.internet import reactor

from pymodbus.client.asynchronous import schedulers
from pymodbus.client.asynchronous.serial import AsyncModbusSerialClient
from pymodbus.client.asynchronous.twisted import ModbusClientProtocol


log = logging.getLogger("pymodbus")
log.setLevel(logging.DEBUG)

# ---------------------------------------------------------------------------#
# state a few constants
# ---------------------------------------------------------------------------#

SERIAL_PORT = "/tmp/ptyp0"  # nosec
STATUS_REGS = (1, 2)
STATUS_COILS = (1, 3)
CLIENT_DELAY = 1
UNIT = 0x01


class ExampleProtocol(ModbusClientProtocol):
    """Example protocol."""

    def __init__(self, framer):
        """Initialize our custom protocol

        :param framer: The decoder to use to process messages
        :param endpoint: The endpoint to send results to
        """
        ModbusClientProtocol.__init__(self, framer)
        log.debug("Beginning the processing loop")
        reactor.callLater(CLIENT_DELAY, self.fetch_holding_registers)

    def fetch_holding_registers(self):
        """Defer fetching holding registers"""
        log.debug("Starting the next cycle")
        result = self.read_holding_registers(*STATUS_REGS, unit=UNIT)
        result.addCallbacks(self.send_holding_registers, self.error_handler)

    def send_holding_registers(self, response):
        """Write values of holding registers, defer fetching coils

        :param response: The response to process
        """
        log.info(response.getRegister(0))
        log.info(response.getRegister(1))
        result = self.read_coils(*STATUS_COILS, unit=UNIT)
        result.addCallbacks(self.start_next_cycle, self.error_handler)

    def start_next_cycle(self, response):
        """Write values of coils, trigger next cycle

        :param response: The response to process
        """
        log.info(response.getBit(0))
        log.info(response.getBit(1))
        log.info(response.getBit(2))
        reactor.callLater(CLIENT_DELAY, self.fetch_holding_registers)

    def error_handler(self, failure):
        """Handle any twisted errors

        :param failure: The error to handle
        """
        log.error(failure)


if __name__ == "__main__":
    import time

    proto, client = AsyncModbusSerialClient(
        schedulers.REACTOR,
        method="rtu",
        port=SERIAL_PORT,
        timeout=2,
        proto_cls=ExampleProtocol,
    )
    proto.start()
    time.sleep(10)  # Wait for operation to complete
    # proto.stop()



#!/usr/bin/env python3
# pylint: disable=missing-type-doc
"""Doxygen API Builder."""
import os
import shutil


def is_exe(path):
    """Return if the program is executable.

    :param path: The path to the file
    :return: True if it is, False otherwise
    """
    return os.path.exists(path) and os.access(path, os.X_OK)


def which(program):
    """Check to see if an executable exists.

    :param program: The program to check for
    :return: The full path of the executable or None if not found
    """
    fpath, _ = os.path.split(program)
    if fpath:
        if is_exe(program):
            return program
    else:
        for path in os.environ["PATH"].split(os.pathsep):
            exe_file = os.path.join(path, program)
            if is_exe(exe_file):
                return exe_file
    return None


if which("doxygen") is not None:
    print("Building Doxygen API Documentation")
    os.system("doxygen .doxygen")  # nosec
    if os.path.exists("../../../build"):
        shutil.move("html", "../../../build/doxygen")
else:
    print("Doxygen not available...not building")



"""Test utilities."""
import struct
import unittest

from pymodbus.utilities import (
    checkCRC,
    checkLRC,
    default,
    dict_property,
    pack_bitstring,
    unpack_bitstring,
)


_test_master = {4: "d"}


class DictPropertyTester:  # pylint: disable=too-few-public-methods
    """Dictionary property test."""

    def __init__(self):
        """Initialize."""
        self.test = {1: "a"}
        self._test = {2: "b"}
        self.__test = {3: "c"}  # pylint: disable=unused-private-member

    l_1 = dict_property(lambda s: s.test, 1)
    l_2 = dict_property(lambda s: s._test, 2)  # pylint: disable=protected-access
    l_3 = dict_property(lambda s: s.__test, 3)  # pylint: disable=protected-access
    s_1 = dict_property("test", 1)
    s_2 = dict_property("_test", 2)
    g_1 = dict_property(_test_master, 4)


class SimpleUtilityTest(unittest.TestCase):
    """Unittest for the pymod.utilities module."""

    def setUp(self):
        """Initialize the test environment"""
        self.data = struct.pack(">HHHH", 0x1234, 0x2345, 0x3456, 0x4567)
        self.string = b"test the computation"
        self.bits = [True, False, True, False, True, False, True, False]

    def tearDown(self):
        """Clean up the test environment"""
        del self.bits
        del self.string

    def test_dict_property(self):
        """Test all string <=> bit packing functions"""
        result = DictPropertyTester()
        self.assertEqual(result.l_1, "a")
        self.assertEqual(result.l_2, "b")
        self.assertEqual(result.l_3, "c")
        self.assertEqual(result.s_1, "a")
        self.assertEqual(result.s_2, "b")
        self.assertEqual(result.g_1, "d")

        for store in "l_1 l_2 l_3 s_1 s_2 g_1".split(" "):
            setattr(result, store, "x")

        self.assertEqual(result.l_1, "x")
        self.assertEqual(result.l_2, "x")
        self.assertEqual(result.l_3, "x")
        self.assertEqual(result.s_1, "x")
        self.assertEqual(result.s_2, "x")
        self.assertEqual(result.g_1, "x")

    def test_default_value(self):
        """Test all string <=> bit packing functions"""
        self.assertEqual(default(1), 0)
        self.assertEqual(default(1.1), 0.0)
        self.assertEqual(default(1 + 1j), 0j)
        self.assertEqual(default("string"), "")
        self.assertEqual(default([1, 2, 3]), [])
        self.assertEqual(default({1: 1}), {})
        self.assertEqual(default(True), False)

    def test_bit_packing(self):
        """Test all string <=> bit packing functions"""
        self.assertEqual(unpack_bitstring(b"\x55"), self.bits)
        self.assertEqual(pack_bitstring(self.bits), b"\x55")

    def test_longitudinal_redundancycheck(self):
        """Test the longitudinal redundancy check code"""
        self.assertTrue(checkLRC(self.data, 0x1C))
        self.assertTrue(checkLRC(self.string, 0x0C))

    def test_cyclic_redundancy_check(self):
        """Test the cyclic redundancy check code"""
        self.assertTrue(checkCRC(self.data, 0xE2DB))
        self.assertTrue(checkCRC(self.string, 0x889E))



"""Pymodbus Client Payload Example.

This example shows how to build a client with a
complicated memory layout using builder-


Works out of the box together with payload_server.py
"""
from collections import OrderedDict
import logging

from pymodbus.client import ModbusTcpClient as ModbusClient
from pymodbus.constants import Endian
from pymodbus.payload import BinaryPayloadBuilder, BinaryPayloadDecoder


# --------------------------------------------------------------------------- #
# configure the client logging
# --------------------------------------------------------------------------- #

FORMAT = (
    "%(asctime)-15s %(threadName)-15s"
    " %(levelname)-8s %(module)-15s:%(lineno)-8s %(message)s"
)
logging.basicConfig(format=FORMAT)
log = logging.getLogger()
log.setLevel(logging.INFO)

ORDER_DICT = {"<": "LITTLE", ">": "BIG"}


def run_binary_payload_client():
    """Run binary payload."""
    # ----------------------------------------------------------------------- #
    # We are going to use a simple sync client to send our requests
    # ----------------------------------------------------------------------- #
    client = ModbusClient("127.0.0.1", port=5020)
    client.connect()

    # ----------------------------------------------------------------------- #
    # If you need to build a complex message to send, you can use the payload
    # builder to simplify the packing logic
    #
    # Packing/unpacking depends on your CPU´s word/byte order. Modbus messages
    # are always using big endian. BinaryPayloadBuilder will pr default use
    # what your CPU uses.
    # The wordorder is applicable only for 32 and 64 bit values
    # Lets say we need to write a value 0x12345678 to a 32 bit register
    # The following combinations could be used to write the register
    # ++++++++++++++++++++++++++++++++++++++++++++ #
    # Word Order  | Byte order | Word1  | Word2  |
    # ------------+------------+--------+--------+
    #     Big     |     Big    | 0x1234 | 0x5678 |
    #     Big     |    Little  | 0x3412 | 0x7856 |
    #    Little   |     Big    | 0x5678 | 0x1234 |
    #    Little   |    Little  | 0x7856 | 0x3412 |
    # ++++++++++++++++++++++++++++++++++++++++++++ #

    # ----------------------------------------------------------------------- #
    for word_endian, byte_endian in (
        (Endian.Big, Endian.Big),
        (Endian.Big, Endian.Little),
        (Endian.Little, Endian.Big),
        (Endian.Little, Endian.Little),
    ):
        print("-" * 60)
        print(f"Word Order: {ORDER_DICT[word_endian]}")
        print(f"Byte Order: {ORDER_DICT[byte_endian]}")
        print()
        builder = BinaryPayloadBuilder(
            wordorder=word_endian,
            byteorder=byte_endian,
        )
        # Normally just do:  builder = BinaryPayloadBuilder()
        my_string = "abcdefgh"
        builder.add_string(my_string)
        builder.add_bits([0, 1, 0, 1, 1, 0, 1, 0])
        builder.add_8bit_int(-0x12)
        builder.add_8bit_uint(0x12)
        builder.add_16bit_int(-0x5678)
        builder.add_16bit_uint(0x1234)
        builder.add_32bit_int(-0x1234)
        builder.add_32bit_uint(0x12345678)
        builder.add_16bit_float(12.34)
        builder.add_16bit_float(-12.34)
        builder.add_32bit_float(22.34)
        builder.add_32bit_float(-22.34)
        builder.add_64bit_int(-0xDEADBEEF)
        builder.add_64bit_uint(0x12345678DEADBEEF)
        builder.add_64bit_uint(0x12345678DEADBEEF)
        builder.add_64bit_float(123.45)
        builder.add_64bit_float(-123.45)
        registers = builder.to_registers()
        print("Writing Registers:")
        print(registers)
        print("\n")
        payload = builder.build()
        address = 0
        # We can write registers
        client.write_registers(address, registers, unit=1)
        # Or we can write an encoded binary string
        client.write_registers(address, payload, skip_encode=True, unit=1)

        # ----------------------------------------------------------------------- #
        # If you need to decode a collection of registers in a weird layout, the
        # payload decoder can help you as well.
        # ----------------------------------------------------------------------- #
        print("Reading Registers:")
        address = 0x0
        count = len(payload)
        result = client.read_holding_registers(address, count, slave=1)
        print(result.registers)
        print("\n")
        decoder = BinaryPayloadDecoder.fromRegisters(
            result.registers, byteorder=byte_endian, wordorder=word_endian
        )
        # Make sure word/byte order is consistent between BinaryPayloadBuilder and BinaryPayloadDecoder
        assert (
            decoder._byteorder == builder._byteorder  # pylint: disable=protected-access
        )  # nosec
        assert (
            decoder._wordorder == builder._wordorder  # pylint: disable=protected-access
        )  # nosec

        decoded = OrderedDict(
            [
                ("string", decoder.decode_string(len(my_string))),
                ("bits", decoder.decode_bits()),
                ("8int", decoder.decode_8bit_int()),
                ("8uint", decoder.decode_8bit_uint()),
                ("16int", decoder.decode_16bit_int()),
                ("16uint", decoder.decode_16bit_uint()),
                ("32int", decoder.decode_32bit_int()),
                ("32uint", decoder.decode_32bit_uint()),
                ("16float", decoder.decode_16bit_float()),
                ("16float2", decoder.decode_16bit_float()),
                ("32float", decoder.decode_32bit_float()),
                ("32float2", decoder.decode_32bit_float()),
                ("64int", decoder.decode_64bit_int()),
                ("64uint", decoder.decode_64bit_uint()),
                ("ignore", decoder.skip_bytes(8)),
                ("64float", decoder.decode_64bit_float()),
                ("64float2", decoder.decode_64bit_float()),
            ]
        )
        print("Decoded Data")
        for name, value in iter(decoded.items()):
            print(
                "%s\t" % name,  # pylint: disable=consider-using-f-string
                hex(value) if isinstance(value, int) else value,
            )
        print("\n")

    # ----------------------------------------------------------------------- #
    # close the client
    # ----------------------------------------------------------------------- #
    client.close()


if __name__ == "__main__":
    run_binary_payload_client()



"""Context for datastore."""
# pylint: disable=missing-type-doc
import logging

from pymodbus.constants import Defaults
from pymodbus.datastore.store import ModbusSequentialDataBlock
from pymodbus.exceptions import NoSuchSlaveException
from pymodbus.interfaces import IModbusSlaveContext


# ---------------------------------------------------------------------------#
#  Logging
# ---------------------------------------------------------------------------#
_logger = logging.getLogger(__name__)


# ---------------------------------------------------------------------------#
#  Slave Contexts
# ---------------------------------------------------------------------------#
class ModbusSlaveContext(IModbusSlaveContext):
    """This creates a modbus data model with each data access stored in a block."""

    def __init__(self, *args, **kwargs):  # pylint: disable=unused-argument
        """Initialize the datastores.

        Defaults to fully populated
        sequential data blocks if none are passed in.

        :param kwargs: Each element is a ModbusDataBlock

            "di" - Discrete Inputs initializer
            "co" - Coils initializer
            "hr" - Holding Register initializer
            "ir" - Input Registers iniatializer
        """
        self.store = {}
        self.store["d"] = kwargs.get("di", ModbusSequentialDataBlock.create())
        self.store["c"] = kwargs.get("co", ModbusSequentialDataBlock.create())
        self.store["i"] = kwargs.get("ir", ModbusSequentialDataBlock.create())
        self.store["h"] = kwargs.get("hr", ModbusSequentialDataBlock.create())
        self.zero_mode = kwargs.get("zero_mode", Defaults.ZeroMode)

    def __str__(self):
        """Return a string representation of the context.

        :returns: A string representation of the context
        """
        return "Modbus Slave Context"

    def reset(self):
        """Reset all the datastores to their default values."""
        for datastore in iter(self.store.values()):
            datastore.reset()

    def validate(self, fc_as_hex, address, count=1):
        """Validate the request to make sure it is in range.

        :param fc_as_hex: The function we are working with
        :param address: The starting address
        :param count: The number of values to test
        :returns: True if the request in within range, False otherwise
        """
        if not self.zero_mode:
            address = address + 1
        txt = f"validate: fc-[{fc_as_hex}] address-{address}: count-{count}"
        _logger.debug(txt)
        return self.store[self.decode(fc_as_hex)].validate(address, count)

    def getValues(self, fc_as_hex, address, count=1):
        """Get `count` values from datastore.

        :param fc_as_hex: The function we are working with
        :param address: The starting address
        :param count: The number of values to retrieve
        :returns: The requested values from a:a+c
        """
        if not self.zero_mode:
            address = address + 1
        txt = f"getValues: fc-[{fc_as_hex}] address-{address}: count-{count}"
        _logger.debug(txt)
        return self.store[self.decode(fc_as_hex)].getValues(address, count)

    def setValues(self, fc_as_hex, address, values):
        """Set the datastore with the supplied values.

        :param fc_as_hex: The function we are working with
        :param address: The starting address
        :param values: The new values to be set
        """
        if not self.zero_mode:
            address = address + 1
        txt = f"setValues[{fc_as_hex}] address-{address}: count-{len(values)}"
        _logger.debug(txt)
        self.store[self.decode(fc_as_hex)].setValues(address, values)

    def register(self, function_code, fc_as_hex, datablock=None):
        """Register a datablock with the slave context.

        :param function_code: function code (int)
        :param fc_as_hex: string representation of function code (e.g "cf" )
        :param datablock: datablock to associate with this function code
        """
        self.store[fc_as_hex] = datablock or ModbusSequentialDataBlock.create()
        self._IModbusSlaveContext__fx_mapper[  # pylint: disable=no-member
            function_code
        ] = fc_as_hex


class ModbusServerContext:
    """This represents a master collection of slave contexts.

    If single is set to true, it will be treated as a single
    context so every unit-id returns the same context. If single
    is set to false, it will be interpreted as a collection of
    slave contexts.
    """

    def __init__(self, slaves=None, single=True):
        """Initialize a new instance of a modbus server context.

        :param slaves: A dictionary of client contexts
        :param single: Set to true to treat this as a single context
        """
        self.single = single
        self._slaves = slaves or {}
        if self.single:
            self._slaves = {Defaults.Slave: self._slaves}

    def __iter__(self):
        """Iterate over the current collection of slave contexts.

        :returns: An iterator over the slave contexts
        """
        return iter(self._slaves.items())

    def __contains__(self, slave):
        """Check if the given slave is in this list.

        :param slave: slave The slave to check for existence
        :returns: True if the slave exists, False otherwise
        """
        if self.single and self._slaves:
            return True
        return slave in self._slaves

    def __setitem__(self, slave, context):
        """Use to set a new slave context.

        :param slave: The slave context to set
        :param context: The new context to set for this slave
        :raises NoSuchSlaveException:
        """
        if self.single:
            slave = Defaults.Slave
        if 0xF7 >= slave >= 0x00:
            self._slaves[slave] = context
        else:
            raise NoSuchSlaveException(f"slave index :{slave} out of range")

    def __delitem__(self, slave):
        """Use to access the slave context.

        :param slave: The slave context to remove
        :raises NoSuchSlaveException:
        """
        if not self.single and (0xF7 >= slave >= 0x00):
            del self._slaves[slave]
        else:
            raise NoSuchSlaveException(f"slave index: {slave} out of range")

    def __getitem__(self, slave):
        """Use to get access to a slave context.

        :param slave: The slave context to get
        :returns: The requested slave context
        :raises NoSuchSlaveException:
        """
        if self.single:
            slave = Defaults.Slave
        if slave in self._slaves:
            return self._slaves.get(slave)
        raise NoSuchSlaveException(
            f"slave - {slave} does not exist, or is out of range"
        )

    def slaves(self):
        """Define slaves."""
        # Python3 now returns keys() as iterable
        return list(self._slaves.keys())



#!/usr/bin/env python3
"""Pymodbus Asynchronous Client Examples.

The following is an example of how to use the asynchronous modbus
client implementation from pymodbus using Tornado.
"""
import functools
import logging

from tornado.ioloop import IOLoop

from pymodbus.client.asynchronous import schedulers

# from pymodbus.client.asynchronous.udp import AsyncModbusUDPClient as ModbusClient
from pymodbus.client.asynchronous.tcp import (
    AsyncModbusTCPClient as ModbusClient,
)


# ---------------------------------------------------------------------------#
# choose the requested modbus protocol
# ---------------------------------------------------------------------------#


# ---------------------------------------------------------------------------#
# configure the client logging
# ---------------------------------------------------------------------------#
_logger = logging.getLogger()
_logger.setLevel(logging.DEBUG)


# ---------------------------------------------------------------------------#
# helper method to test deferred callbacks
# ---------------------------------------------------------------------------#


def dassert(future, callback):
    """Dassert."""

    def _assertor(value):
        # by pass assertion, an error here stops the write callbacks
        assert value  # nosec

    def on_done(f_trans):
        if exc := f_trans.exception():
            _logger.debug(exc)
            return _assertor(False)

        return _assertor(callback(f_trans.result()))

    future.add_done_callback(on_done)


def _print(value):
    """Print."""
    if hasattr(value, "bits"):
        result = value.bits
    elif hasattr(value, "registers"):
        result = value.registers
    else:
        _logger.error(value)
        return None
    txt = f"Printing : -- {result}"
    _logger.info(txt)
    return result


UNIT = 0x01

# ---------------------------------------------------------------------------#
# example requests
# ---------------------------------------------------------------------------#
# simply call the methods that you would like to use. An example session
# is displayed below along with some assert checks. Note that unlike the
# synchronous version of the client, the asynchronous version returns
# deferreds which can be thought of as a handle to the callback to send
# the result of the operation.  We are handling the result using the
# deferred assert helper(dassert).
# ---------------------------------------------------------------------------#


def begin_asynchronous_test(client, protocol):
    """Begin async test."""
    rq = client.write_coil(1, True, unit=UNIT)
    rr = client.read_coils(1, 1, unit=UNIT)
    dassert(rq, lambda r: r.function_code < 0x80)  # test for no error
    dassert(rr, _print)  # test the expected value

    rq = client.write_coils(1, [False] * 8, unit=UNIT)
    rr = client.read_coils(1, 8, unit=UNIT)
    dassert(rq, lambda r: r.function_code < 0x80)  # test for no error
    dassert(rr, _print)  # test the expected value

    rq = client.write_coils(1, [False] * 8, unit=UNIT)
    rr = client.read_discrete_inputs(1, 8, unit=UNIT)
    dassert(rq, lambda r: r.function_code < 0x80)  # test for no error
    dassert(rr, _print)  # test the expected value

    rq = client.write_register(1, 10, unit=UNIT)
    rr = client.read_holding_registers(1, 1, unit=UNIT)
    dassert(rq, lambda r: r.function_code < 0x80)  # test for no error
    dassert(rr, _print)  # test the expected value

    rq = client.write_registers(1, [10] * 8, unit=UNIT)
    rr = client.read_input_registers(1, 8, unit=UNIT)
    dassert(rq, lambda r: r.function_code < 0x80)  # test for no error
    dassert(rr, _print)  # test the expected value

    arguments = {
        "read_address": 1,
        "read_count": 8,
        "write_address": 1,
        "write_registers": [20] * 8,
    }
    rq = client.readwrite_registers(**arguments, unit=UNIT)
    rr = client.read_input_registers(1, 8, unit=UNIT)
    dassert(rq, lambda r: r.registers == [20] * 8)  # test the expected value
    dassert(rr, _print)  # test the expected value

    # -----------------------------------------------------------------------#
    # close the client at some time later
    # -----------------------------------------------------------------------#
    IOLoop.current().add_timeout(IOLoop.current().time() + 1, client.close)
    IOLoop.current().add_timeout(IOLoop.current().time() + 2, protocol.stop)


# ---------------------------------------------------------------------------#
# choose the client you want
# ---------------------------------------------------------------------------#
# make sure to start an implementation to hit against. For this
# you can use an existing device, the reference implementation in the tools
# directory, or start a pymodbus server.
# ---------------------------------------------------------------------------#


def err(*args, **kwargs):
    """Error."""
    txt = f"Err {args} {kwargs}"
    _logger.error(txt)


def callback(protocol, future):
    """Call as callback."""
    _logger.debug("Client connected")
    if exp := future.exception():
        return err(exp)

    client = future.result()
    return begin_asynchronous_test(client, protocol)


if __name__ == "__main__":
    protocol, future = ModbusClient(schedulers.IO_LOOP, port=5020)
    future.add_done_callback(functools.partial(callback, protocol))



#!/usr/bin/env python3
"""Pymodbus synchronous forwarder.

This is a repeater or converter and  an example of just how powerful datastore is.

It consist of a server (any comm) and a client (any comm) and basically all request
received by the server is sent by client and all responses received by the
client is sent back by the server.

Both server and client are tcp based, but it can be easily modified to any server/client
(see client_sync.py and server_sync.py for other communication types)
"""
import argparse
import logging

from pymodbus.client import ModbusTcpClient
from pymodbus.datastore import ModbusServerContext
from pymodbus.datastore.remote import RemoteSlaveContext
from pymodbus.server import StartTcpServer


def run_forwarder():
    """Run forwarder setup."""
    port_server, port_client, slaves = get_commandline()

    client = ModbusTcpClient(
        host="localhost",
        port=port_client,
    )

    # If required to communicate with a specified client use unit=<unit_id>
    # in RemoteSlaveContext
    # For e.g to forward the requests to slave with unit address 1 use
    # store = RemoteSlaveContext(client, unit=1)
    if slaves:
        store = {}
        for i in slaves:
            store[i.to_bytes(1, "big")] = RemoteSlaveContext(client, unit=i)
    else:
        store = RemoteSlaveContext(client)
    context = ModbusServerContext(slaves=store, single=True)

    # start forwarding client and server
    client.connect()
    StartTcpServer(context=context, address=("localhost", port_server))
    # loop forever


# --------------------------------------------------------------------------- #
# Extra code, to allow commandline parameters instead of changing the code
# --------------------------------------------------------------------------- #
FORMAT = "%(asctime)-15s %(levelname)-8s %(module)-15s:%(lineno)-8s %(message)s"
logging.basicConfig(format=FORMAT)
_logger = logging.getLogger()


def get_commandline():
    """Read and validate command line arguments"""
    parser = argparse.ArgumentParser(description="Command line options for examples")
    parser.add_argument(
        "--log",
        choices=["critical", "error", "warning", "info", "debug"],
        help='"critical", "error", "warning", "info" or "debug"',
        type=str,
    )
    parser.add_argument(
        "--port",
        help="the port to use",
        type=int,
    )
    parser.add_argument(
        "--port_client",
        help="the port to use",
        type=int,
    )
    parser.add_argument(
        "--slaves",
        help="list of slaves to forward",
        type=int,
        nargs="+",
    )
    args = parser.parse_args()

    # set defaults
    _logger.setLevel(args.log.upper() if args.log else logging.INFO)
    if not args.port:
        args.port = 5020
    if not args.port_client:
        args.port_client = 5010

    return args.port, args.port_client, args.slaves


if __name__ == "__main__":
    run_forwarder()



"""Test client sync."""
from pymodbus.repl.client.main import _process_args as process_args
from pymodbus.server.reactive.default_config import DEFAULT_CONFIG


def test_repl_default_config():
    """Test default config can be loaded."""
    config = DEFAULT_CONFIG
    assert config is not None


def test_repl_client_process_args():
    """Test argument processing in repl.client.main (_process_args function)."""
    resp = process_args(["address=11"], False)
    assert resp == ({"address": 11}, True)

    resp = process_args(["address=0x11"], False)
    assert resp == ({"address": 17}, True)

    resp = process_args(["address=0b11"], False)
    assert resp == ({"address": 3}, True)

    resp = process_args(["address=0o11"], False)
    assert resp == ({"address": 9}, True)

    resp = process_args(["address=11", "value=0x10"], False)
    assert resp == ({"address": 11, "value": 16}, True)

    resp = process_args(["value=11", "address=0x10"], False)
    assert resp == ({"address": 16, "value": 11}, True)

    resp = process_args(["address=0b11", "value=0x10"], False)
    assert resp == ({"address": 3, "value": 16}, True)

    try:
        resp = process_args(["address=0xhj", "value=0x10"], False)
    except ValueError:
        pass

    try:
        resp = process_args(["address=11ah", "value=0x10"], False)
    except ValueError:
        pass

    try:
        resp = process_args(["address=0b12", "value=0x10"], False)
    except ValueError:
        pass



"""Command Completion for pymodbus REPL."""
from prompt_toolkit.application.current import get_app

# pylint: disable=missing-type-doc
from prompt_toolkit.completion import Completer, Completion
from prompt_toolkit.filters import Condition
from prompt_toolkit.styles import Style

from pymodbus.repl.client.helper import get_commands


@Condition
def has_selected_completion():
    """Check for selected completion."""
    complete_state = get_app().current_buffer.complete_state
    return complete_state is not None and complete_state.current_completion is not None


style = Style.from_dict(
    {
        "completion-menu.completion": "bg:#008888 #ffffff",
        "completion-menu.completion.current": "bg:#00aaaa #000000",
        "scrollbar.background": "bg:#88aaaa",
        "scrollbar.button": "bg:#222222",
    }
)


class CmdCompleter(Completer):
    """Completer for Pymodbus REPL."""

    def __init__(self, client=None, commands=None, ignore_case=True):
        """Initialize.

        :param client: Modbus Client
        :param commands: Commands to be added for Completion (list)
        :param ignore_case: Ignore Case while looking up for commands
        """
        self._commands = commands or get_commands(client)
        self._commands["help"] = ""
        self._command_names = self._commands.keys()
        self.ignore_case = ignore_case

    @property
    def commands(self):
        """Return commands."""
        return self._commands

    @property
    def command_names(self):
        """Return command names."""
        return self._commands.keys()

    def completing_command(self, words, word_before_cursor):
        """Determine if we are dealing with supported command.

        :param words: Input text broken in to word tokens.
        :param word_before_cursor: The current word before the cursor, \
            which might be one or more blank spaces.
        :return:
        """
        return len(words) == 1 and len(word_before_cursor)

    def completing_arg(self, words, word_before_cursor):
        """Determine if we are currently completing an argument.

        :param words: The input text broken into word tokens.
        :param word_before_cursor: The current word before the cursor, \
            which might be one or more blank spaces.
        :return: Specifies whether we are currently completing an arg.
        """
        return len(words) > 1 and len(word_before_cursor)

    def arg_completions(
        self, words, word_before_cursor
    ):  # pylint: disable=unused-argument
        """Generate arguments completions based on the input.

        :param words: The input text broken into word tokens.
        :param word_before_cursor: The current word  before the cursor, \
            which might be one or more blank spaces.
        :return: A list of completions.
        """
        cmd = words[0].strip()
        cmd = self._commands.get(cmd, None)
        return cmd if cmd else None

    def _get_completions(self, word, word_before_cursor):
        """Get completions."""
        if self.ignore_case:
            word_before_cursor = word_before_cursor.lower()
        return self.word_matches(word, word_before_cursor)

    def word_matches(self, word, word_before_cursor):
        """Match the word and word before cursor.

        :param word: The input text broken into word tokens.
        :param word_before_cursor: The current word before the cursor, \
            which might be one or more blank spaces.
        :return: True if matched.

        """
        if self.ignore_case:
            word = word.lower()
        return word.startswith(word_before_cursor)

    def get_completions(self, document, complete_event):
        """Get completions for the current scope.

        :param document: An instance of `prompt_toolkit.Document`.
        :param complete_event: (Unused).
        :return: Yields an instance of `prompt_toolkit.completion.Completion`.
        """
        word_before_cursor = document.get_word_before_cursor(WORD=True)
        text = document.text_before_cursor.lstrip()
        words = document.text.strip().split()
        meta = None
        commands = []
        if not words:
            # yield commands
            pass
        if self.completing_command(words, word_before_cursor):
            commands = self._command_names
            c_meta = {
                k: v.help_text if not isinstance(v, str) else v
                for k, v in self._commands.items()
            }
            meta = lambda x: (  # pylint: disable=unnecessary-lambda-assignment
                x,
                c_meta.get(x, ""),
            )
        else:
            if not list(
                filter(lambda cmd: any(x == cmd for x in words), self._command_names)
            ):
                # yield commands
                pass

            if " " in text:
                command = self.arg_completions(words, word_before_cursor)
                commands = list(command.get_completion())
                commands = list(
                    filter(lambda cmd: not (any(cmd in x for x in words)), commands)
                )
                meta = command.get_meta
        for command in commands:
            if self._get_completions(command, word_before_cursor):
                _, display_meta = meta(command) if meta else ("", "")
                yield Completion(
                    command, -len(word_before_cursor), display_meta=display_meta
                )



"""Modbus Server Datastore.

For each server, you will create a ModbusServerContext and pass
in the default address space for each data access.  The class
will create and manage the data.

Further modification of said data accesses should be performed
with [get,set][access]Values(address, count)

Datastore Implementation
-------------------------

There are two ways that the server datastore can be implemented.
The first is a complete range from "address" start to "count"
number of indices.  This can be thought of as a straight array::

    data = range(1, 1 + count)
    [1,2,3,...,count]

The other way that the datastore can be implemented (and how
many devices implement it) is a associate-array::

    data = {1:"1", 3:"3", ..., count:"count"}
    [1,3,...,count]

The difference between the two is that the latter will allow
arbitrary gaps in its datastore while the former will not.
This is seen quite commonly in some modbus implementations.
What follows is a clear example from the field:

Say a company makes two devices to monitor power usage on a rack.
One works with three-phase and the other with a single phase. The
company will dictate a modbus data mapping such that registers::

    n:      phase 1 power
    n+1:    phase 2 power
    n+2:    phase 3 power

Using this, layout, the first device will implement n, n+1, and n+2,
however, the second device may set the latter two values to 0 or
will simply not implemented the registers thus causing a single read
or a range read to fail.

I have both methods implemented, and leave it up to the user to change
based on their preference.
"""
# pylint: disable=missing-type-doc
import logging

from pymodbus.exceptions import NotImplementedException, ParameterException


# ---------------------------------------------------------------------------#
#  Logging
# ---------------------------------------------------------------------------#
_logger = logging.getLogger(__name__)


# ---------------------------------------------------------------------------#
#  Datablock Storage
# ---------------------------------------------------------------------------#
class BaseModbusDataBlock:
    """Base class for a modbus datastore

    Derived classes must create the following fields:
            @address The starting address point
            @defult_value The default value of the datastore
            @values The actual datastore values

    Derived classes must implemented the following methods:
            validate(self, address, count=1)
            getValues(self, address, count=1)
            setValues(self, address, values)
    """

    def default(self, count, value=False):
        """Use to initialize a store to one value.

        :param count: The number of fields to set
        :param value: The default value to set to the fields
        """
        self.default_value = value  # pylint: disable=attribute-defined-outside-init
        self.values = [  # pylint: disable=attribute-defined-outside-init
            self.default_value
        ] * count
        self.address = 0x00  # pylint: disable=attribute-defined-outside-init

    def reset(self):
        """Reset the datastore to the initialized default value."""
        self.values = [  # pylint: disable=attribute-defined-outside-init
            self.default_value
        ] * len(self.values)

    def validate(self, address, count=1):
        """Check to see if the request is in range.

        :param address: The starting address
        :param count: The number of values to test for
        :raises NotImplementedException:
        """
        raise NotImplementedException("Datastore Address Check")

    def getValues(self, address, count=1):  # pylint: disable=invalid-name
        """Return the requested values from the datastore.

        :param address: The starting address
        :param count: The number of values to retrieve
        :raises NotImplementedException:
        """
        raise NotImplementedException("Datastore Value Retrieve")

    def setValues(self, address, values):  # pylint: disable=invalid-name
        """Return the requested values from the datastore.

        :param address: The starting address
        :param values: The values to store
        :raises NotImplementedException:
        """
        raise NotImplementedException("Datastore Value Retrieve")

    def __str__(self):
        """Build a representation of the datastore.

        :returns: A string representation of the datastore
        """
        return f"DataStore({len(self.values)}, {self.default_value})"

    def __iter__(self):
        """Iterate over the data block data.

        :returns: An iterator of the data block data
        """
        if isinstance(self.values, dict):
            return iter(self.values.items())
        return enumerate(self.values, self.address)


class ModbusSequentialDataBlock(BaseModbusDataBlock):
    """Creates a sequential modbus datastore."""

    def __init__(self, address, values):
        """Initialize the datastore.

        :param address: The starting address of the datastore
        :param values: Either a list or a dictionary of values
        """
        self.address = address
        if hasattr(values, "__iter__"):
            self.values = list(values)
        else:
            self.values = [values]
        self.default_value = self.values[0].__class__()

    @classmethod
    def create(cls):
        """Create a datastore.

        With the full address space initialized to 0x00

        :returns: An initialized datastore
        """
        return cls(0x00, [0x00] * 65536)

    def validate(self, address, count=1):
        """Check to see if the request is in range.

        :param address: The starting address
        :param count: The number of values to test for
        :returns: True if the request in within range, False otherwise
        """
        result = self.address <= address
        result &= (self.address + len(self.values)) >= (address + count)
        return result

    def getValues(self, address, count=1):
        """Return the requested values of the datastore.

        :param address: The starting address
        :param count: The number of values to retrieve
        :returns: The requested values from a:a+c
        """
        start = address - self.address
        return self.values[start : start + count]

    def setValues(self, address, values):
        """Set the requested values of the datastore.

        :param address: The starting address
        :param values: The new values to be set
        """
        if not isinstance(values, list):
            values = [values]
        start = address - self.address
        self.values[start : start + len(values)] = values


class ModbusSparseDataBlock(BaseModbusDataBlock):
    """Create a sparse modbus datastore.

    E.g Usage.
    sparse = ModbusSparseDataBlock({10: [3, 5, 6, 8], 30: 1, 40: [0]*20})

    This would create a datablock with 3 blocks starting at
    offset 10 with length 4 , 30 with length 1 and 40 with length 20

    sparse = ModbusSparseDataBlock([10]*100)
    Creates a sparse datablock of length 100 starting at offset 0 and default value of 10

    sparse = ModbusSparseDataBlock() --> Create Empty datablock
    sparse.setValues(0, [10]*10)  --> Add block 1 at offset 0 with length 10 (default value 10)
    sparse.setValues(30, [20]*5)  --> Add block 2 at offset 30 with length 5 (default value 20)

    if mutable is set to True during initialization, the datablock can not be altered with
    setValues (new datablocks can not be added)
    """

    def __init__(self, values=None, mutable=True):
        """Initialize a sparse datastore.

        Will only answer to addresses
        registered, either initially here, or later via setValues()

        :param values: Either a list or a dictionary of values
        :param mutable: The data-block can be altered later with setValues(i.e add more blocks)

        If values are list , This is as good as sequential datablock.
        Values as dictionary should be in {offset: <values>} format, if values
        is a list, a sparse datablock is created starting at offset with the length of values.
        If values is a integer, then the value is set for the corresponding offset.

        """
        self.values = {}
        self._process_values(values)
        self.mutable = mutable
        self.default_value = self.values.copy()
        self.address = next(iter(self.values.keys()), None)

    @classmethod
    def create(cls, values=None):
        """Create sparse datastore.

        Use setValues to initialize registers.

        :param values: Either a list or a dictionary of values
        :returns: An initialized datastore
        """
        return cls(values)

    def reset(self):
        """Reset the store to the initially provided defaults."""
        self.values = self.default_value.copy()

    def validate(self, address, count=1):
        """Check to see if the request is in range.

        :param address: The starting address
        :param count: The number of values to test for
        :returns: True if the request in within range, False otherwise
        """
        if not count:
            return False
        handle = set(range(address, address + count))
        return handle.issubset(set(iter(self.values.keys())))

    def getValues(self, address, count=1):
        """Return the requested values of the datastore.

        :param address: The starting address
        :param count: The number of values to retrieve
        :returns: The requested values from a:a+c
        """
        return [self.values[i] for i in range(address, address + count)]

    def _process_values(self, values):
        """Process values."""

        def _process_as_dict(values):
            for idx, val in iter(values.items()):
                if isinstance(val, (list, tuple)):
                    for i, v_item in enumerate(val):
                        self.values[idx + i] = v_item
                else:
                    self.values[idx] = int(val)

        if isinstance(values, dict):
            _process_as_dict(values)
            return
        if hasattr(values, "__iter__"):
            values = dict(enumerate(values))
        elif values is None:
            values = {}  # Must make a new dict here per instance
        else:
            raise ParameterException(
                "Values for datastore must be a list or dictionary"
            )
        _process_as_dict(values)

    def setValues(self, address, values, use_as_default=False):
        """Set the requested values of the datastore.

        :param address: The starting address
        :param values: The new values to be set
        :param use_as_default: Use the values as default
        :raises ParameterException:
        """
        if isinstance(values, dict):
            new_offsets = list(set(values.keys()) - set(self.values.keys()))
            if new_offsets and not self.mutable:
                raise ParameterException(f"Offsets {new_offsets} not in range")
            self._process_values(values)
        else:
            if not isinstance(values, list):
                values = [values]
            for idx, val in enumerate(values):
                if address + idx not in self.values and not self.mutable:
                    raise ParameterException("Offset {address+idx} not in range")
                self.values[address + idx] = val
        if not self.address:
            self.address = next(iter(self.values.keys()), None)
        if use_as_default:
            for idx, val in iter(self.values.items()):
                self.default_value[idx] = val



"""MEI Message Test Fixture.

This fixture tests the functionality of all the
mei based request/response messages:
"""
import unittest

from pymodbus.constants import DeviceInformation
from pymodbus.device import ModbusControlBlock
from pymodbus.mei_message import (
    ReadDeviceInformationRequest,
    ReadDeviceInformationResponse,
)


# ---------------------------------------------------------------------------#
#  Fixture
# ---------------------------------------------------------------------------#

TEST_VERSION = b"v2.1.12"
TEST_MESSAGE = b"\x00\x07Company\x01\x07Product\x02\x07v2.1.12"


class ModbusMeiMessageTest(unittest.TestCase):
    """Unittest for the pymodbus.mei_message module."""

    # -----------------------------------------------------------------------#
    #  Read Device Information
    # -----------------------------------------------------------------------#

    def test_read_device_information_request_encode(self):
        """Test basic bit message encoding/decoding"""
        params = {"read_code": DeviceInformation.Basic, "object_id": 0x00}
        handle = ReadDeviceInformationRequest(**params)
        result = handle.encode()
        self.assertEqual(result, b"\x0e\x01\x00")
        self.assertEqual("ReadDeviceInformationRequest(1,0)", str(handle))

    def test_read_device_information_request_decode(self):
        """Test basic bit message encoding/decoding"""
        handle = ReadDeviceInformationRequest()
        handle.decode(b"\x0e\x01\x00")
        self.assertEqual(handle.read_code, DeviceInformation.Basic)
        self.assertEqual(handle.object_id, 0x00)

    def test_read_device_information_request(self):
        """Test basic bit message encoding/decoding"""
        context = None
        control = ModbusControlBlock()
        control.Identity.VendorName = "Company"
        control.Identity.ProductCode = "Product"
        control.Identity.MajorMinorRevision = TEST_VERSION
        control.Identity.update({0x81: ["Test", "Repeated"]})

        handle = ReadDeviceInformationRequest()
        result = handle.execute(context)
        self.assertTrue(isinstance(result, ReadDeviceInformationResponse))
        self.assertEqual(result.information[0x00], "Company")
        self.assertEqual(result.information[0x01], "Product")
        self.assertEqual(result.information[0x02], TEST_VERSION)
        with self.assertRaises(KeyError):
            _ = result.information[0x81]

        handle = ReadDeviceInformationRequest(
            read_code=DeviceInformation.Extended, object_id=0x80
        )
        result = handle.execute(context)
        self.assertEqual(result.information[0x81], ["Test", "Repeated"])

    def test_read_device_information_request_error(self):
        """Test basic bit message encoding/decoding"""
        handle = ReadDeviceInformationRequest()
        handle.read_code = -1
        self.assertEqual(handle.execute(None).function_code, 0xAB)
        handle.read_code = 0x05
        self.assertEqual(handle.execute(None).function_code, 0xAB)
        handle.object_id = -1
        self.assertEqual(handle.execute(None).function_code, 0xAB)
        handle.object_id = 0x100
        self.assertEqual(handle.execute(None).function_code, 0xAB)

    def test_read_device_information_encode(self):
        """Test that the read fifo queue response can encode"""
        message = b"\x0e\x01\x83\x00\x00\x03"
        message += TEST_MESSAGE
        dataset = {
            0x00: "Company",
            0x01: "Product",
            0x02: TEST_VERSION,
        }
        handle = ReadDeviceInformationResponse(
            read_code=DeviceInformation.Basic, information=dataset
        )
        result = handle.encode()
        self.assertEqual(result, message)
        self.assertEqual("ReadDeviceInformationResponse(1)", str(handle))

        dataset = {
            0x00: "Company",
            0x01: "Product",
            0x02: TEST_VERSION,
            0x81: ["Test", "Repeated"],
        }
        message = b"\x0e\x03\x83\x00\x00\x05"
        message += TEST_MESSAGE
        message += b"\x81\x04Test\x81\x08Repeated"
        handle = ReadDeviceInformationResponse(
            read_code=DeviceInformation.Extended, information=dataset
        )
        result = handle.encode()
        self.assertEqual(result, message)

    def test_read_device_information_encode_long(self):
        """Test that the read fifo queue response can encode"""
        longstring = (
            "Lorem ipsum dolor sit amet, consectetur adipiscing "
            "elit. Vivamus rhoncus massa turpis, sit amet ultrices"
            " orci semper ut. Aliquam tristique sapien in lacus "
            "pharetra, in convallis nunc consectetur. Nunc velit "
            "elit, vehicula tempus tempus sed. "
        )

        message = b"\x0e\x01\x83\xFF\x80\x03"
        message += TEST_MESSAGE
        dataset = {
            0x00: "Company",
            0x01: "Product",
            0x02: TEST_VERSION,
            0x80: longstring,
        }
        handle = ReadDeviceInformationResponse(
            read_code=DeviceInformation.Basic, information=dataset
        )
        result = handle.encode()
        self.assertEqual(result, message)
        self.assertEqual("ReadDeviceInformationResponse(1)", str(handle))

    def test_read_device_information_decode(self):
        """Test that the read device information response can decode"""
        message = b"\x0e\x01\x01\x00\x00\x05"
        message += TEST_MESSAGE
        message += b"\x81\x04Test\x81\x08Repeated\x81\x07Another"
        handle = ReadDeviceInformationResponse(read_code=0x00, information=[])
        handle.decode(message)
        self.assertEqual(handle.read_code, DeviceInformation.Basic)
        self.assertEqual(handle.conformity, 0x01)
        self.assertEqual(handle.information[0x00], b"Company")
        self.assertEqual(handle.information[0x01], b"Product")
        self.assertEqual(handle.information[0x02], TEST_VERSION)
        self.assertEqual(handle.information[0x81], [b"Test", b"Repeated", b"Another"])

    def test_rtu_frame_size(self):
        """Test that the read device information response can decode"""
        message = (
            b"\x04\x2B\x0E\x01\x81\x00\x01\x01\x00\x06\x66\x6F\x6F\x62\x61\x72\xD7\x3B"
        )
        result = ReadDeviceInformationResponse.calculateRtuFrameSize(message)
        self.assertEqual(result, 18)
        message = b"\x00\x2B\x0E\x02\x00\x4D\x47"
        result = ReadDeviceInformationRequest.calculateRtuFrameSize(message)
        self.assertEqual(result, 7)



"""Pymodbus REPL Entry point."""
# pylint: disable=anomalous-backslash-in-string
# flake8: noqa: W605
import logging
import pathlib
import sys


try:
    import click
except ImportError:
    print('click not installed!! Install with "pip install click"')
    sys.exit(1)
try:
    from prompt_toolkit import PromptSession, print_formatted_text
except ImportError:
    print(
        "prompt toolkit is not installed!! "
        'Install with "pip install prompt_toolkit --upgrade"'
    )
    sys.exit(1)

from prompt_toolkit.auto_suggest import AutoSuggestFromHistory
from prompt_toolkit.formatted_text import HTML
from prompt_toolkit.history import FileHistory
from prompt_toolkit.key_binding import KeyBindings
from prompt_toolkit.lexers import PygmentsLexer
from prompt_toolkit.styles import Style
from pygments.lexers.python import PythonLexer

from pymodbus.exceptions import ParameterException
from pymodbus.repl.client.completer import CmdCompleter, has_selected_completion
from pymodbus.repl.client.helper import CLIENT_ATTRIBUTES, Result
from pymodbus.repl.client.mclient import ModbusSerialClient, ModbusTcpClient
from pymodbus.transaction import (
    ModbusAsciiFramer,
    ModbusBinaryFramer,
    ModbusRtuFramer,
    ModbusSocketFramer,
)
from pymodbus.version import version


_logger = logging.getLogger()

click.disable_unicode_literals_warning = True

TITLE = f"""
----------------------------------------------------------------------------
__________          _____             .___  __________              .__
\______   \___.__. /     \   ____   __| _/  \______   \ ____ ______ |  |
 |     ___<   |  |/  \ /  \ /  _ \ / __ |    |       _// __ \\\____ \|  |
 |    |    \___  /    Y    (  <_> ) /_/ |    |    |   \  ___/|  |_> >  |__
 |____|    / ____\____|__  /\____/\____ | /\ |____|_  /\___  >   __/|____/
           \/            \/            \/ \/        \/     \/|__|
                                        v1.3.0 - {version}
----------------------------------------------------------------------------
"""


style = Style.from_dict(
    {
        "completion-menu.completion": "bg:#008888 #ffffff",
        "completion-menu.completion.current": "bg:#00aaaa #000000",
        "scrollbar.background": "bg:#88aaaa",
        "scrollbar.button": "bg:#222222",
    }
)


def bottom_toolbar():
    """Do console toolbar.

    :return:
    """
    return HTML(
        'Press <b><style bg="ansired">CTRL+D or exit </style></b>'
        ' to exit! Type "help" for list of available commands'
    )


class CaseInsenstiveChoice(click.Choice):
    """Do case Insensitive choice for click commands and options."""

    def convert(self, value, param, ctx):
        """Convert args to uppercase for evaluation."""
        if value is None:
            return None
        return super().convert(value.strip().upper(), param, ctx)


class NumericChoice(click.Choice):
    """Do numeric choice for click arguments and options."""

    def __init__(self, choices, typ):
        """Initialize."""
        self.typ = typ
        super().__init__(choices)

    def convert(self, value, param, ctx):
        """Convert."""
        # Exact match
        if value in self.choices:
            return self.typ(value)

        if ctx is not None and ctx.token_normalize_func is not None:
            value = ctx.token_normalize_func(value)
            for choice in self.casted_choices:  # pylint: disable=no-member
                if ctx.token_normalize_func(choice) == value:
                    return choice

        self.fail(
            "invalid choice: %s. (choose from %s)"  # pylint: disable=consider-using-f-string
            % (
                value,
                ", ".join(self.choices),
            ),
            param,
            ctx,
        )
        return None


def _process_args(args: list, string: bool=True):
    """Internal function to parse arguments provided on command line.

    :param args: Array of argument values
    :param string: True if arguments values are strings, false if argument values are integers

    :return Tuple, where the first member is hash of parsed values, and second is boolean flag
        indicating if parsing succeeded.
    """
    kwargs = {}
    execute = True
    skip_index = None

    def _parse_val(arg_name, val):
        if not string:
            if "," in val:
                val = val.split(",")
                val = [int(v, 0) for v in val]
            else:
                val = int(val, 0)
        kwargs[arg_name] = val

    for i, arg in enumerate(args):
        if i == skip_index:
            continue
        arg = arg.strip()
        if "=" in arg:
            arg_name, val = arg.split("=")
            _parse_val(arg_name, val)
        else:
            arg_name, val = arg, args[i + 1]
            try:
                _parse_val(arg_name, val)
                skip_index = i + 1
            except TypeError:
                click.secho("Error parsing arguments!", fg="yellow")
                execute = False
                break
            except ValueError:
                click.secho("Error parsing argument", fg="yellow")
                execute = False
                break
    return kwargs, execute


def cli(client):  # noqa: C901 pylint: disable=too-complex
    """Run client definition."""
    use_keys = KeyBindings()
    history_file = pathlib.Path.home().joinpath(".pymodhis")

    @use_keys.add("c-space")
    def _(event):
        """Initialize autocompletion, or select the next completion."""
        buff = event.app.current_buffer
        if buff.complete_state:
            buff.complete_next()
        else:
            buff.start_completion(select_first=False)

    @use_keys.add("enter", filter=has_selected_completion)
    def _(event):
        """Make the enter key work as the tab key only when showing the menu."""
        event.current_buffer.complete_state = None
        buffer = event.cli.current_buffer
        buffer.complete_state = None

    session = PromptSession(
        lexer=PygmentsLexer(PythonLexer),
        completer=CmdCompleter(client),
        style=style,
        complete_while_typing=True,
        bottom_toolbar=bottom_toolbar,
        key_bindings=use_keys,
        history=FileHistory(history_file),
        auto_suggest=AutoSuggestFromHistory(),
    )
    click.secho(TITLE, fg="green")
    result = None
    while True:  # pylint: disable=too-many-nested-blocks
        try:

            text = session.prompt("> ", complete_while_typing=True)
            if text.strip().lower() == "help":
                print_formatted_text(HTML("<u>Available commands:</u>"))
                for cmd, obj in sorted(session.completer.commands.items()):
                    if cmd != "help":
                        print_formatted_text(
                            HTML(
                                "<skyblue>{:45s}</skyblue>"  # pylint: disable=consider-using-f-string
                                "<seagreen>{:100s}"
                                "</seagreen>".format(cmd, obj.help_text)
                            )
                        )

                continue
            if text.strip().lower() == "exit":
                raise EOFError()
            if text.strip().lower().startswith("client."):
                try:
                    text = text.strip().split()
                    cmd = text[0].split(".")[1]
                    args = text[1:]
                    kwargs, execute = _process_args(args, string=False)
                    if execute:
                        if text[0] in CLIENT_ATTRIBUTES:
                            result = Result(getattr(client, cmd))
                        else:
                            result = Result(getattr(client, cmd)(**kwargs))
                        result.print_result()
                except Exception as exc:  # pylint: disable=broad-except
                    click.secho(repr(exc), fg="red")
            elif text.strip().lower().startswith("result."):
                if result:
                    words = text.lower().split()
                    if words[0] == "result.raw":
                        result.raw()
                    if words[0] == "result.decode":
                        args = words[1:]
                        kwargs, execute = _process_args(args)
                        if execute:
                            result.decode(**kwargs)
        except KeyboardInterrupt:
            continue  # Control-C pressed. Try again.
        except EOFError:
            break  # Control-D pressed.
        except Exception as exc:  # Handle all other exceptions pylint: disable=broad-except
            click.secho(str(exc), fg="red")

    click.secho("GoodBye!", fg="blue")


@click.group("pymodbus-repl")
@click.version_option(version, message=TITLE)
@click.option("--verbose", is_flag=True, default=False, help="Verbose logs")
@click.option(
    "--broadcast-support",
    is_flag=True,
    default=False,
    help="Support broadcast messages",
)
@click.option(
    "--retry-on-empty", is_flag=True, default=False, help="Retry on empty response"
)
@click.option(
    "--retry-on-error", is_flag=True, default=False, help="Retry on error response"
)
@click.option("--retries", default=3, help="Retry count")
@click.option(
    "--reset-socket/--no-reset-socket",
    default=True,
    help="Reset client socket on error",
)
@click.pass_context
def main(
    ctx,
    verbose,
    broadcast_support,
    retry_on_empty,
    retry_on_error,
    retries,
    reset_socket,
):
    """Run Main."""
    if verbose:
        use_format = (
            "%(asctime)-15s %(threadName)-15s "
            "%(levelname)-8s %(module)-15s:%(lineno)-8s %(message)s"
        )
        logging.basicConfig(format=use_format)
        _logger.setLevel(logging.DEBUG)
    ctx.obj = {
        "broadcast": broadcast_support,
        "retry_on_empty": retry_on_empty,
        "retry_on_invalid": retry_on_error,
        "retries": retries,
        "reset_socket": reset_socket,
    }


@main.command("tcp")
@click.pass_context
@click.option("--host", default="localhost", help="Modbus TCP IP ")
@click.option(
    "--port",
    default=502,
    type=int,
    help="Modbus TCP port",
)
@click.option(
    "--framer",
    default="tcp",
    type=str,
    help="Override the default packet framer tcp|rtu",
)
def tcp(ctx, host, port, framer):
    """Define TCP."""
    kwargs = {"host": host, "port": port}
    kwargs.update(**ctx.obj)
    if framer == "rtu":
        kwargs["framer"] = ModbusRtuFramer
    client = ModbusTcpClient(**kwargs)
    cli(client)


@main.command("serial")
@click.pass_context
@click.option(
    "--method",
    default="rtu",
    type=str,
    help="Modbus Serial Mode (rtu/ascii)",
)
@click.option(
    "--port",
    default=None,
    type=str,
    help="Modbus RTU port",
)
@click.option(
    "--baudrate",
    help="Modbus RTU serial baudrate to use. Defaults to 9600",
    default=9600,
    type=int,
)
@click.option(
    "--bytesize",
    help="Modbus RTU serial Number of data bits. "
    "Possible values: FIVEBITS, SIXBITS, SEVENBITS, "
    "EIGHTBITS. Defaults to 8",
    type=NumericChoice(["5", "6", "7", "8"], int),
    default="8",
)
@click.option(
    "--parity",
    help="Modbus RTU serial parity. "
    " Enable parity checking. Possible values: "
    "PARITY_NONE, PARITY_EVEN, PARITY_ODD PARITY_MARK, "
    'PARITY_SPACE. Default to "N"',
    default="N",
    type=CaseInsenstiveChoice(["N", "E", "O", "M", "S"]),
)
@click.option(
    "--stopbits",
    help="Modbus RTU serial stop bits. "
    "Number of stop bits. Possible values: STOPBITS_ONE, "
    'STOPBITS_ONE_POINT_FIVE, STOPBITS_TWO. Default to "1"',
    default="1",
    type=NumericChoice(["1", "1.5", "2"], float),
)
@click.option(
    "--xonxoff",
    help="Modbus RTU serial xonxoff.  Enable software flow control. Defaults to 0",
    default=0,
    type=int,
)
@click.option(
    "--rtscts",
    help="Modbus RTU serial rtscts. Enable hardware (RTS/CTS) flow "
    "control. Defaults to 0",
    default=0,
    type=int,
)
@click.option(
    "--dsrdtr",
    help="Modbus RTU serial dsrdtr. Enable hardware (DSR/DTR) flow "
    "control. Defaults to 0",
    default=0,
    type=int,
)
@click.option(
    "--timeout",
    help="Modbus RTU serial read timeout. Defaults to 0.025 sec",
    default=0.25,
    type=float,
)
@click.option(
    "--write-timeout",
    help="Modbus RTU serial write timeout. Defaults to 2 sec",
    default=2,
    type=float,
)
def serial(  # pylint: disable=too-many-arguments
    ctx,
    method,
    port,
    baudrate,
    bytesize,
    parity,
    stopbits,
    xonxoff,
    rtscts,
    dsrdtr,
    timeout,
    write_timeout,
):
    """Define serial communication."""
    method = method.lower()
    if method == "ascii":
        framer = ModbusAsciiFramer
    elif method == "rtu":
        framer = ModbusRtuFramer
    elif method == "binary":
        framer = ModbusBinaryFramer
    elif method == "socket":
        framer = ModbusSocketFramer
    else:
        raise ParameterException("Invalid framer method requested")
    client = ModbusSerialClient(
        framer=framer,
        port=port,
        baudrate=baudrate,
        bytesize=bytesize,
        parity=parity,
        stopbits=stopbits,
        xonxoff=xonxoff,
        rtscts=rtscts,
        dsrdtr=dsrdtr,
        timeout=timeout,
        write_timeout=write_timeout,
        **ctx.obj,
    )
    cli(client)


if __name__ == "__main__":
    main()  # pylint: disable=no-value-for-parameter



# pylint: disable=missing-type-doc
r"""This is used to generate decoder blocks.

so that non-programmers can define the
register values and then decode a modbus device all
without having to write a line of code for decoding.

Currently supported formats are:

* csv
* json
* xml

Here is an example of generating and using a mapping decoder
(note that this is still in the works and will be greatly
simplified in the final api; it is just an example of the
requested functionality)::

    CSV:
    address,type,size,name,function
    0,int16,1,Comm. count PLC,hr
    1,int16,1,Comm. count PLC,hr
    2,int16,1,Comm. count PLC,hr
    3,int16,1,Comm. count PLC,hr
    4,int16,1,Comm. count PLC,hr
    5,int16,1,Comm. count PLC,hr
    6,int16,1,Comm. count PLC,hr
    7,int16,1,Comm. count PLC,hr
    8,int16,1,Comm. count PLC,hr
    9,int16,1,Comm. count PLC,hr
    10,int32,2,Comm. count PLC,hr
    12,int32,2,Comm. count PLC,hr

    from modbus_mapper import csv_mapping_parser
    from modbus_mapper import mapping_decoder
    from pymodbus.client import ModbusTcpClient
    from pymodbus.payload import BinaryPayloadDecoder
    from pymodbus.constants import Endian

    from pprint import pprint
    import logging

    # FORMAT = "%(asctime)-15s %(levelname)-8s %(module)-15s:%(lineno)-8s %(message)s"
    # logging.basicConfig(format=FORMAT)
    # _logger = logging.getLogger()
    # _logger.setLevel(logging.DEBUG)

    template = ["address", "type", "size", "name", "function"]
    raw_mapping = csv_mapping_parser("simple_mapping_client.csv", template)
    # raw_mapping = csv_mapping_parser("Naust_Comm_to_scr_client.csv", template)
    mapping = mapping_decoder(raw_mapping)

    client = ModbusTcpClient(host="localhost", port=5020)

    response = client.read_holding_registers(address=int(0), count=14)
    decoder = BinaryPayloadDecoder.fromRegisters(
        response.registers, byteorder=Endian.Big, wordorder=Endian.Little
    )

    for block in mapping.items():
        for mac in block:
            if type(mac) == dict:
                # response = client.read_holding_registers(
                #     address=int(mac["address"]), count=mac["size"]
                # )
                # decoder = BinaryPayloadDecoder.fromRegisters(
                #     response.registers, byteorder=Endian.Big, wordorder=Endian.Little
                # )
                print("[{}]\t{}".format(mac["address"], mac["type"]()(decoder)))
                # decoder._payload # remove mac["size"] bytes from beginning

Also, using the same input mapping parsers, we can generate
populated slave contexts that can be run behind a modbus server::

    CSV:
    address,value,function,name,description
    0,0,hr,Comm. count PLC,Comm. count PLC
    1,10,hr,Comm. count PLC,Comm. count PLC
    2,20,hr,Comm. count PLC,Comm. count PLC
    3,30,hr,Comm. count PLC,Comm. count PLC
    4,40,hr,Comm. count PLC,Comm. count PLC
    5,50,hr,Comm. count PLC,Comm. count PLC
    6,60,hr,Comm. count PLC,Comm. count PLC
    7,70,hr,Comm. count PLC,Comm. count PLC
    8,80,hr,Comm. count PLC,Comm. count PLC
    9,90,hr,Comm. count PLC,Comm. count PLC
    10,100,hr,Comm. count PLC,Comm. count PLC
    11,0,hr,Comm. count PLC,Comm. count PLC
    12,120,hr,Comm. count PLC,Comm. count PLC
    13,0,hr,Comm. count PLC,Comm. count PLC

    from modbus_mapper import csv_mapping_parser
    from modbus_mapper import modbus_context_decoder

    from pymodbus.server import StartTcpServer
    from pymodbus.datastore.context import ModbusServerContext
    from pymodbus.device import ModbusDeviceIdentification
    from pymodbus.version import version


    from pprint import pprint
    import logging

    FORMAT = "%(asctime)-15s %(levelname)-8s %(module)-15s:%(lineno)-8s %(message)s"
    logging.basicConfig(format=FORMAT)
    _logger = logging.getLogger()
    _logger.setLevel(logging.DEBUG)

    template = ["address", "value", "function", "name", "description"]
    raw_mapping = csv_mapping_parser("simple_mapping_server.csv", template)

    slave_context = modbus_context_decoder(raw_mapping)
    context = ModbusServerContext(slaves=slave_context, single=True)
    identity = ModbusDeviceIdentification(
        info_name={
            "VendorName": "Pymodbus",
            "ProductCode": "PM",
            "VendorUrl": "https://github.com/riptideio/pymodbus/",
            "ProductName": "Pymodbus Server",
            "ModelName": "Pymodbus Server",
            "MajorMinorRevision": version.short(),
        }
    )
    StartTcpServer(context=context, identity=identity, address=("localhost", 5020))

"""
from collections import defaultdict
import csv
from io import StringIO
import json
from tokenize import generate_tokens

from pymodbus.datastore import (
    ModbusSlaveContext,
    ModbusSparseDataBlock,
)

# --------------------------------------------------------------------------- #
# raw mapping input parsers
# --------------------------------------------------------------------------- #
# These generate the raw mapping_blocks from some form of input
# which can then be passed to the decoder in question to supply
# the requested output result.
# --------------------------------------------------------------------------- #


def csv_mapping_parser(path, template):
    """Given a csv file of the the mapping data for a modbus device,

    return a mapping layout that can be used to decode an new block.

    .. note:: For the template, a few values are required
    to be defined: address, size, function, and type. All the remaining
    values will be stored, but not formatted by the application.
    So for example::

        template = ["address", "type", "size", "name", "function"]
        mappings = json_mapping_parser("mapping.json", template)

    :param path: The path to the csv input file
    :param template: The row value template
    :returns: The decoded csv dictionary
    """
    mapping_blocks = defaultdict(dict)
    with open(path, "r") as handle:  # pylint: disable=unspecified-encoding
        reader = csv.reader(handle)
        next(reader)  # skip the csv header
        for row in reader:
            mapping = dict(zip(template, row))
            # mapping.pop("function")
            aid = mapping["address"]
            mapping_blocks[aid] = mapping
    return mapping_blocks


def json_mapping_parser(path, template):
    """Given a json file of the the mapping data for a modbus device,

    return a mapping layout that can
    be used to decode an new block.

    .. note:: For the template, a few values are required
    to be mapped: address, size, and type. All the remaining
    values will be stored, but not formatted by the application.
    So for example::

        template = {
            "Start": "address",
            "DataType": "type",
            "Length": "size"
            # the remaining keys will just pass through
        }
        mappings = json_mapping_parser("mapping.json", template)

    :param path: The path to the csv input file
    :param template: The row value template
    :returns: The decoded csv dictionary
    """
    mapping_blocks = {}
    with open(path, "r") as handle:  # pylint: disable=unspecified-encoding
        for tid, rows in json.load(handle).iteritems():
            mappings = {}
            for key, values in rows.iteritems():
                mapping = {template.get(k, k): v for k, v in values.iteritems()}
                mappings[int(key)] = mapping
            mapping_blocks[tid] = mappings
    return mapping_blocks


def xml_mapping_parser():
    """Given an xml file of the the mapping data for a modbus device,

    return a mapping layout that can be used to decode an new block.

    :returns: The decoded csv dictionary
    """


# --------------------------------------------------------------------------- #
# modbus context decoders
# --------------------------------------------------------------------------- #
# These are used to decode a raw mapping_block into a slave context with
# populated function data blocks.
# --------------------------------------------------------------------------- #
def modbus_context_decoder(mapping_blocks):
    """Generate a backing slave context with initialized data blocks.

    .. note:: This expects the following for each block:
    address, value, and function where function is one of
    di (discretes), co (coils), hr (holding registers), or
    ir (input registers).

    :param mapping_blocks: The mapping blocks
    :returns: The initialized modbus slave context
    """
    sparse = ModbusSparseDataBlock()
    sparse.create()
    for block in mapping_blocks.items():
        for mapping in block:
            if type(mapping) == dict:
                value = mapping["value"]
                address = mapping["address"]
                sparse.setValues(address=int(address), values=int(value))
    return ModbusSlaveContext(
        di=sparse, co=sparse, hr=sparse, ir=sparse, zero_mode=True
    )


# --------------------------------------------------------------------------- #
# modbus mapping decoder
# --------------------------------------------------------------------------- #
# These are used to decode a raw mapping_block into a request decoder.
# So this allows one to simply grab a number of registers, and then
# pass them to this decoder which will do the rest.
# --------------------------------------------------------------------------- #
class ModbusTypeDecoder:
    """This is a utility to determine the correct decoder to use given a type name.

    By default this supports all the types available in the default modbus
    decoder, however this can easily be extended this class
    and adding new types to the mapper::

        class CustomTypeDecoder(ModbusTypeDecoder):
            def __init__(self):
                ModbusTypeDecode.__init__(self)
                self.mapper["type-token"] = self.callback

            def parse_my_bitfield(self, tokens):
                return lambda d: d.decode_my_type()

    """

    def __init__(self):
        """Initialize a new instance of the decoder"""
        self.default = lambda m: self.parse_16bit_uint
        self.parsers = {
            "uint": self.parse_16bit_uint,
            "uint8": self.parse_8bit_uint,
            "uint16": self.parse_16bit_uint,
            "uint32": self.parse_32bit_uint,
            "uint64": self.parse_64bit_uint,
            "int": self.parse_16bit_int,
            "int8": self.parse_8bit_int,
            "int16": self.parse_16bit_int,
            "int32": self.parse_32bit_int,
            "int64": self.parse_64bit_int,
            "float": self.parse_32bit_float,
            "float32": self.parse_32bit_float,
            "float64": self.parse_64bit_float,
            "string": self.parse_32bit_int,
            "bits": self.parse_bits,
        }

    # ------------------------------------------------------------ #
    # Type parsers
    # ------------------------------------------------------------ #
    @staticmethod
    def parse_string(tokens):
        """Parse value."""
        _ = next(tokens)
        size = int(next(tokens))
        return lambda d: d.decode_string(size=size)

    @staticmethod
    def parse_bits():
        """Parse value."""
        return lambda d: d.decode_bits()

    @staticmethod
    def parse_8bit_uint():
        """Parse value."""
        return lambda d: d.decode_8bit_uint()

    @staticmethod
    def parse_16bit_uint():
        """Parse value."""
        return lambda d: d.decode_16bit_uint()

    @staticmethod
    def parse_32bit_uint():
        """Parse value."""
        return lambda d: d.decode_32bit_uint()

    @staticmethod
    def parse_64bit_uint():
        """Parse value."""
        return lambda d: d.decode_64bit_uint()

    @staticmethod
    def parse_8bit_int():
        """Parse value."""
        return lambda d: d.decode_8bit_int()

    @staticmethod
    def parse_16bit_int():
        """Parse value."""
        return lambda d: d.decode_16bit_int()

    @staticmethod
    def parse_32bit_int():
        """Parse value."""
        return lambda d: d.decode_32bit_int()

    @staticmethod
    def parse_64bit_int():
        """Parse value."""
        return lambda d: d.decode_64bit_int()

    @staticmethod
    def parse_32bit_float():
        """Parse value."""
        return lambda d: d.decode_32bit_float()

    @staticmethod
    def parse_64bit_float():
        """Parse value."""
        return lambda d: d.decode_64bit_float()

    # ------------------------------------------------------------
    # Public Interface
    # ------------------------------------------------------------
    def tokenize(self, value):
        """Return the tokens

        :param value: The value to tokenize
        :returns: A token generator
        """
        tokens = generate_tokens(StringIO(value).readline)
        for _, tokval, _, _, _ in tokens:
            yield tokval

    def parse(self, value):
        """Return a function that supplied with a decoder,

        will decode the correct value.

        :param value: The type of value to parse
        :returns: The decoder method to use
        """
        tokens = self.tokenize(value)
        token = next(tokens).lower()  # pylint: disable=no-member
        parser = self.parsers.get(token, self.default)
        return parser


def mapping_decoder(mapping_blocks, decoder=None):
    """Convert them into modbus value decoder map.

    :param mapping_blocks: The mapping blocks
    :param decoder: The type decoder to use
    """
    decoder = decoder or ModbusTypeDecoder()
    map = defaultdict(dict)
    for block in mapping_blocks.items():
        for mapping in block:
            if type(mapping) == dict:
                mapping["address"] = mapping["address"]
                mapping["size"] = mapping["size"]
                mapping["type"] = decoder.parse(mapping["type"])
                map[mapping["address"]] = mapping
    return map
